#! /usr/bin/env python
# spumux.py

"""This module is for adding subtitles to MPEG video files using spumux.

Defined here are two functions for adding subtitles to an MPEG file:

    add_subpictures:  Add image files (.png)
    add_subtitles:    Add subtitle files (.sub, .srt etc.)

Use these if you just want to add subpictures or subtitles, and don't want
to think much about the XML internals.
"""

__all__ = [\
    'add_subpictures',
    'add_subtitles']

from libtovid import utils
from libtovid import xml

# spumux XML elements and valid attributes
"""
subpictures
stream
textsub
    ['filename',
    'characterset',
    'fontsize',
    'font',
    'horizontal-alignment',
    'vertical-alignment',
    'left-margin',
    'right-margin',
    'subtitle-fps',
    'movie-fps',
    'movie-width',
    'movie-height']
button
    ['name',
    'x0', 'y0', # Upper-left corner, inclusively
    'x1', 'y1', # Lower-right corner, exclusively
    'up', 'down', 'left', 'right']
action
    ['name']
spu
    ['start',
    'end',
    'image',
    'highlight',
    'select',
    'transparent',   # color code
    'force',         # 'yes' (force display, required for menus)
    'autooutline',   # 'infer'
    'outlinewidth',
    'autoorder',     # 'rows' or 'columns'
    'xoffset',
    'yoffset']
"""

###
### Internal functions
###

def get_xml(textsub_or_spu):
    subpictures = xml.Element('subpictures')
    stream = subpictures.add('stream')
    stream.add_child(textsub_or_spu)
    return str(subpictures)


def get_xmlfile(textsub_or_spu):
    """Write spumux XML file for the given Textsub or Spu element, and
    return the written filename.
    """
    xmldata = get_xml(textsub_or_spu)
    xmlfile = utils.temp_file(suffix=".xml")
    xmlfile.write(xmldata)
    xmlfile.close()
    return xmlfile.name


def mux_subs(subtitle, movie_filename, stream_id=0):
    """Run spumux to multiplex the given subtitle with an .mpg file.
    
        subtitle:       Textsub or Spu element
        movie_filename: Name of an .mpg file to multiplex subtitle into
        stream_id:      Stream ID number to pass to spumux
    """
    # Create temporary .mpg file in the same directory
    base_dir = os.path.dirname(movie_filename)
    subbed_filename = temp_name(suffix=".mpg", dir=base_dir)
    # spumux xmlfile < movie_filename > subbed_filename
    spumux = Command('spumux',
                     '-s%s' % stream_id,
                     xmlfile.name)
    spumux.run_redir(movie_filename, subbed_filename)
    spumux.wait()

    # Remove old file
    os.remove(movie_filename)
    # Rename temporary file to new file
    os.rename(subbed_filename, movie_filename)
    # Remove the XML file
    os.remove(xmlfile.name)


###
### Exported functions
###

def add_subpictures(movie_filename, select, image=None, highlight=None):
    """Adds PNG image subpictures to an .mpg video file to create a DVD menu.
    
        select:    Image shown as the navigational selector or "cursor"
        image:     Image shown for non-selected regions
        highlight: Image shown when "enter" is pressed
        
    All images must be indexed, 4-color, transparent, non-antialiased PNG.
    Button regions are auto-inferred.
    """
    spu = xml.Element('spu')
    spu.set(start='0',
            force='yes',
            select=select,
            image=image,
            highlight=highlight)
    # TODO Find a good default outlinewidth
    spu.set(autooutline=infer, outlinewidth=10)
    # TODO: Support explicit button regions
    mux_subs(spu, movie_filename)


def add_subtitles(movie_filename, sub_filenames):
    """Adds one or more subtitle files to an .mpg video file.
    
        movie_filename: Name of .mpg file to add subtitles to
        sub_filenames:  Filename or list of filenames of subtitle
                        files to include (.sub/.srt etc.)

    """
    infile = load_media(movie_filename)
    width, height = infile.scale

    # Convert sub_filenames to list if necessary
    if type(sub_filenames) == str:
        sub_filenames = [sub_filenames]
    # spumux each subtitle file with its own stream ID
    for stream_id, sub_filename in enumerate(sub_filenames):
        # <textsub attribute="value" .../>
        textsub = xml.Element('textsub')
        textsub.set(movie_fps=infile.fps,
                    movie_width=width,
                    movie_height=height,
                    filename=sub_filename)
        mux_subs(textsub, movie_filename, stream_id)


if __name__ == '__main__':
    print "spumux XML examples"

    print "Subpicture example:"
    spu = xml.Element('spu')
    spu.add('button', name='but1', down='but2')
    spu.add('button', name='but2', up='but1')
    print get_xml(spu)

    print "Text subtitle example:"
    textsub = xml.Element('textsub')
    textsub.set(filename='foo.sub',
                fontsize=14.0,
                font="Luxi Mono")
    print get_xml(textsub)



#! /usr/bin/env python
# animation.py

"""This module provides classes and functions for working with animation.
Two classes are provided:

    Keyframe: A frame with a specific data value
    Tween:    A data sequence interpolated from Keyframes

The data being interpolated may represent color, opacity, location, or anything
else that can be described numerically. Keyframe data may be scalar (single
integers or decimal values) or vector (tuples such as (x, y) coordinates or
(r, g, b) color values).

For example, let's define three keyframes:

    >>> keys = [Keyframe(1, 0),
    ...         Keyframe(6, 50),
    ...         Keyframe(12, 10)]

The value increases from 0 to 50 over frames 1-6, then back down to 10
over frames 6-12. The values at intermediate frames (2-5 and 7-11) can be
interpolated or "tweened" automatically, using the Tween class:

    >>> tween = Tween(keys)
    >>> tween.data
    [0, 10, 20, 30, 40, 50, 43, 36, 30, 23, 16, 10]

Another example using tweening of (x, y) coordinates:

    >>> keys = [Keyframe(1, (20, 20)),
    ...         Keyframe(6, (80, 20)),
    ...         Keyframe(12, (100, 100))]

Here, a point on a two-dimensional plane starts at (20, 20), moving first
to the right, to (80, 20), then diagonally to (100, 100).

    >>> tween = Tween(keys)
    >>> for (x, y) in tween.data:
    ...     print (x, y)
    ...
    (20, 20)
    (32, 20)
    (44, 20)
    (56, 20)
    (68, 20)
    (80, 20)
    (83, 33)
    (86, 46)
    (90, 60)
    (93, 73)
    (96, 86)
    (100, 100)

"""

__all__ = [\
    'Keyframe',
    'lerp',
    'cos_interp',
    'interpolate',
    'tween']

import copy
import doctest
import math

class Keyframe:
    """Associates a specific frame in an animation with a numeric value.
    A Keyframe is a (frame, data) pair defining a "control point" on a graph:
    
            100 |
                |       Keyframe(10, 50)
           data |      *
                |
              0 |__________________________
                1     10     20     30
                        frame
    
    The data can represent anything you like. For instance, opacity:
    
            100 |* Keyframe(1, 100)
                |       
     opacity(%) |
                |
              0 |____________________* Keyframe(30, 0)
                1     10     20     30
                        frame

    See the Tween class below for what you can do with these Keyframes,
    once you have them.
    """
    def __init__(self, frame, data):
        """Create a keyframe, associating a given frame with some data.
        The data may be an integer, floating-point value, or a tuple like
        (x, y) or (r, g, b).
        """
        self.frame = frame
        self.data = data


# Interpolation algorithms
def lerp(x, (x0, y0), (x1, y1)):
    """Do linear interpolation between points (x0, y0), (x1, y1), and return
    the 'y' of the given 'x'.

    This form of interpolation simply connects two points with a straight
    line. Blunt, but effective."""
    return y0 + (x - x0) * (y1 - y0) / (x1 - x0)

def cos_interp(x, (x0, y0), (x1, y1)):
    """Do cosine-based interpolation between (x0, y0), (x1, y1) and return
    the 'y' of the given 'x'.

    Essentially, a crude alternative to polynomial spline interpolation; this
    method transitions between two values by matching a segment of the cosine
    curve [0, pi] (for decreasing value) or [pi, 2*pi] (for increasing value)
    to the interval between the given points.

    It gives smoother results at inflection points than linear interpolation, 
    but will result in "ripple" effects if keyframes are too dense or many.   
    """
    # Map the interpolation area (domain of x) to [0, pi]
    x_norm = math.pi * (x - x0) / (x1 - x0)
    # For y0 < y1, use upward-sloping part of the cosine curve [pi, 2*pi]
    if y0 < y1:
        x_norm += math.pi
    # Calculate and return resulting y-value
    y_min = min(y1, y0)
    y_diff = abs(y1 - y0)
    return y_min + y_diff * (math.cos(x_norm) + 1) / 2.0


# Single-interval interpolation function
def interpolate(frame, left, right, method):
    """Interpolate data between left and right Keyframes at the given frame,
    using the given interpolation method ('linear' or 'cosine'). Return the
    value at the given frame.
    
    The left and right Keyframes mark the endpoints of the curve to be
    interpolated. For example, if a value changes from 50 to 80 over the
    course of frames 1 to 30:

        >>> left = Keyframe(1, 50)
        >>> right = Keyframe(30, 80)
    
    Then, the value at frame 10 can be interpolated as follows:
    
        >>> interpolate(10, left, right, 'linear')
        59
        >>> interpolate(10, left, right, 'cosine')
        56.582194019564263

    For frames outside the keyframe interval, the corresponding endpoint value
    is returned:
    
        >>> interpolate(0, left, right, 'linear')
        50
        >>> interpolate(40, left, right, 'linear')
        80
    """
    assert isinstance(left, Keyframe) and isinstance(right, Keyframe)
    # At or beyond endpoints, return endpoint value
    if frame <= left.frame:
        return left.data
    elif frame >= right.frame:
        return right.data

    # Use the appropriate interpolation function
    if method == 'cosine':
        interp_func = cos_interp
    else: # method == 'linear'
        interp_func = lerp

    # Interpolate integers or floats
    if isinstance(left.data, int) or isinstance(left.data, float):
        p0 = (left.frame, left.data)
        p1 = (right.frame, right.data)
        return interp_func(frame, p0, p1)
    # Interpolate a tuple (x, y, ...)
    elif isinstance(left.data, tuple):
        # Interpolate each dimension separately
        dim = 0
        result = []
        while dim < len(left.data):
            result.append(interp_func(frame, (left.frame, left.data[dim]), \
                 (right.frame, right.data[dim])))
            dim += 1
        return tuple(result)


class Tween:
    """An "in-between" sequence, calculated by interpolating the data in a
    list of keyframes according to a given interpolation method.
    
    First, define some keyframes:
    
        >>> keyframes = [Keyframe(1, 1), Keyframe(10, 25)]
    
    At frame 1, the value is 1; at frame 10, the value is 25. To get an
    interpolation of the data between these two keyframes, use:
    
        >>> tween = Tween(keyframes)

    Now, retrieve the interpolated data all at once:
    
        >>> tween.data
        [1, 3, 6, 9, 11, 14, 17, 19, 22, 25]
    
    Or by using array notation, indexed by frame number:
    
        >>> tween[3]
        6
        >>> tween[8]
        19
    """
    def __init__(self, keyframes, method='linear'):
        """Create an in-between sequence from a list of keyframes. The
        interpolation method can be 'linear' or 'cosine'.

        See effect.py for implementation examples.
        """
        assert isinstance(keyframes[0], Keyframe)
        self.keyframes = keyframes
        self.method = method
        self.data = []
        self._tween()

    def _tween(self):
        """Perform in-betweening calculation on the current keyframes and
        fill self.data with tweened values, indexed by frame number.
        """
        self.data = []
        # TODO: Sort keyframes in increasing order by frame number (to ensure
        # keyframes[0] is the first frame, and keyframes[-1] is the last frame)
        # Make a copy of keyframes
        keys = copy.deepcopy(self.keyframes)
        self.start = keys[0].frame
        self.end = keys[-1].frame
        first = self.start
        last = self.end
        
        # If keyframe interval is empty, use constant data from first keyframe
        if first == last:
            self.data = keys[0].data
            return
    
        # Pop off keyframes as each interval is calculated
        left = keys.pop(0)
        right = keys.pop(0)
        frame = first
        # Interpolate until the last frame is reached
        while frame <= last:
            value = interpolate(frame, left, right, self.method)
            self.data.append(value)
            # Get the next interval, if it exists
            if frame == right.frame and len(keys) > 0:
                left = right
                right = keys.pop(0)
            frame += 1

    def __getitem__(self, frame):
        """Return the interpolated data at the given frame. This allows
        accessing a tweened value with subscripting by frame number:
        
            >>> keys = [Keyframe(1, 1), Keyframe(30, 50)]
            >>> tween = Tween(keys)
            >>> tween[1]
            1
            >>> tween[30]
            50
            >>> tween[21]
            34

        Frame numbers outside the keyframed region have the same values as
        those at the corresponding endpoints:
        
            >>> tween[0]
            1
            >>> tween[100]
            50
        """
        # If data is a single value, frame is irrelevant
        if not isinstance(self.data, list):
            return self.data
        # Otherwise, return the value for the given frame, extending endpoints
        elif frame < self.start:
            return self.data[0]
        elif frame > self.end:
            return self.data[-1]
        else:
            return self.data[frame - self.start]


if __name__ == '__main__':
    doctest.testmod(verbose=True)

    p0 = (1, 1)
    p1 = (20, 20)
    
    print "Interpolation methods"
    print "x      lerp     cos_interp"
    for x in range(1, 21):
        print "%s      %s       %s" % \
              (x, lerp(x, p0, p1), cos_interp(x, p0, p1))



#! /usr/bin/env python
# dialogs.py

import wx

import libtovid
from libtovid.gui.configs import TovidConfig
from libtovid.gui.constants import *
from libtovid.gui.util import _

__all__ = [\
    "FontChooserDialog",
    "PreferencesDialog"]

class FontChooserDialog(wx.Dialog):
    """A simple font chooser"""

    def __init__(self, parent, id, curFacename = "Default"):
        wx.Dialog.__init__(self, parent, id, _("Font chooser"), wx.DefaultPosition,
                (400, 400))

        # Center dialog
        self.Centre()

        # Get global configuration
        self.curConfig = TovidConfig()

        # Construct a list of fonts. Always list "Default" first.
        strFontChoices = []
        strFontChoices.extend(self.curConfig.wx_IM_FontMap.keys())
        strFontChoices.sort()
        strFontChoices.insert(0, "Default")

        # List of fonts, label and sample of selected font
        self.listFonts = wx.ListBox(self, wx.ID_ANY, wx.DefaultPosition,
                wx.DefaultSize, strFontChoices, wx.LB_SINGLE)
        self.lblFont = wx.StaticText(self, wx.ID_ANY, _("Sample of the selected font:"))
        self.lblFontSample = wx.StaticText(self, wx.ID_ANY, "The quick brown fox\n "
                "jumps over the lazy dog.")
        wx.EVT_LISTBOX(self, self.listFonts.GetId(), self.OnSelectFont)

        # Set listbox selection to given facename
        curFontIdx = self.listFonts.FindString(curFacename)
        self.listFonts.SetSelection(curFontIdx)
        # Remember given facename
        self.font = wx.Font(12, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL,
            wx.FONTWEIGHT_NORMAL, False, "Default")
        # Show sample in current font
        self.lblFontSample.SetFont(self.font)

        # OK/Cancel buttons
        self.btnOK = wx.Button(self, wx.ID_OK, "OK")
        self.btnCancel = wx.Button(self, wx.ID_CANCEL, "Cancel")
        sizButtons = wx.BoxSizer(wx.HORIZONTAL)
        sizButtons.Add(self.btnOK, 1, wx.EXPAND | wx.ALL, 16)
        sizButtons.Add(self.btnCancel, 1, wx.EXPAND | wx.ALL, 16)
        # Sizer to hold controls
        sizMain = wx.BoxSizer(wx.VERTICAL)
        sizMain.Add(self.listFonts, 1, wx.EXPAND | wx.ALL, 8)
        sizMain.Add(self.lblFont, 0, wx.EXPAND | wx.ALL, 8)
        sizMain.Add(self.lblFontSample, 0, wx.EXPAND | wx.ALL, 16)
        sizMain.Add(sizButtons, 0, wx.EXPAND)
        self.SetSizer(sizMain)
        self.Show()

        # If there are very few (<6) fonts available,
        # show a message telling how to get more
        if len(self.curConfig.wx_IM_FontMap.keys()) < 6:
            dlgGetMoreFonts = wx.MessageDialog(self,
                "You have less than six fonts to choose from. See the\n"
                "tovid documentation (http://tovid.org/)\n"
                "for instructions on how to get more.",
                "How to get more fonts", wx.OK | wx.ICON_INFORMATION)
            dlgGetMoreFonts.ShowModal()

    def OnSelectFont(self, evt):
        """Change the sample font to reflect the selected font"""
        face = self.listFonts.GetStringSelection()
        self.font = wx.Font(12, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL,
            wx.FONTWEIGHT_NORMAL, False, face)
        self.lblFontSample.SetFont(self.font)

    def GetSelectedFont(self):
        """Return the font that was selected"""
        return self.font

class PreferencesDialog(wx.Dialog):
    """Preferences/configuration settings dialog. Not yet used."""

    def __init__(self, parent, id):
        wx.Dialog.__init__(self, parent, id, "tovid GUI Preferences", \
                wx.DefaultPosition, (400, 200))
        # Center dialog
        self.Centre()
        # Get global configuration
        self.curConfig = TovidConfig()
        # Heading
        self.txtHeading = HeadingText(self, wx.ID_ANY, "Preferences")
        # OK/Cancel buttons
        self.btnOK = wx.Button(self, wx.ID_OK, "OK")
        self.btnCancel = wx.Button(self, wx.ID_CANCEL, "Cancel")
        self.sizButtons = wx.BoxSizer(wx.HORIZONTAL)
        self.sizButtons.Add(self.btnOK, 1, wx.EXPAND | wx.ALL, 16)
        self.sizButtons.Add(self.btnCancel, 1, wx.EXPAND | wx.ALL, 16)
        wx.EVT_BUTTON(self, wx.ID_OK, self.OnOK)
        # Main sizer to hold all controls
        self.sizMain = wx.BoxSizer(wx.VERTICAL)
        self.sizMain.Add(self.txtHeading, 0, wx.EXPAND | wx.ALL, 8)
        # Controls will go here
        self.sizMain.Add(self.sizButtons, 0, wx.EXPAND | wx.ALL, 8)
        self.SetSizer(self.sizMain)

    def OnOK(self, evt):
        """Assign configuration to underlying Config class"""
        # Config assignment goes here
        self.EndModal(wx.ID_OK)




#! /usr/bin/env python
# odict.py

"""Ordered dictionary class, from a Python Cookbook recipe:

http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/107747

"""
__all__ = ['Odict']

from UserDict import UserDict

class Odict (UserDict):
    def __init__(self, keys=None, values=None):
        """Create an Odict from the given keys and values."""
        keys = keys or []
        values = values or []
        if len(keys) != len(values):
            raise ValueError("keys and values are not the same length")
        self._keys = keys
        UserDict.__init__(self, dict(zip(keys, values)))

    def __delitem__(self, key):
        UserDict.__delitem__(self, key)
        self._keys.remove(key)

    def __setitem__(self, key, item):
        UserDict.__setitem__(self, key, item)
        if key not in self._keys: self._keys.append(key)

    def clear(self):
        UserDict.clear(self)
        self._keys = []

    def copy(self):
        dict = UserDict.copy(self)
        dict._keys = self._keys[:]
        return dict

    def items(self):
        return zip(self._keys, self.values())

    def keys(self):
        return self._keys

    def popitem(self):
        try:
            key = self._keys[-1]
        except IndexError:
            raise KeyError('dictionary is empty')

        val = self[key]
        del self[key]

        return (key, val)

    def setdefault(self, key, failobj = None):
        UserDict.setdefault(self, key, failobj)
        if key not in self._keys: self._keys.append(key)

    def update(self, dict):
        UserDict.update(self, dict)
        for key in dict.keys():
            if key not in self._keys: self._keys.append(key)

    def values(self):
        return map(self.get, self._keys)
    
    def __str__(self):
        lines = []
        for key, value in self.items():
            lines.append("%s: %s" % (key, value))
        return '\n'.join(lines)


def convert_list(choices):
    """Convert a list of choices to an Odict (ordered dictionary).
    choices may be in one of several formats:

               string: 'one|two|three'
                 list: ['one', 'two', 'three']
                 dict: {'a': "Choice A", 'b': "Choice B"}
        list-of-lists: [['a', "Choice A"], ['b', "Choice B"], ..]
    
    Note: the dict form does not preserve order. Use list-of-lists
    to maintain the specified order.
    """
    if type(choices) not in [str, list, dict]:
        raise TypeError("choices must be a string, list, or dictionary.")

    if type(choices) == str:
        choices = choices.split('|')
        return Odict(choices, choices)

    if type(choices) == dict:
        return Odict(choices.keys(), choices.values())

    # choices is a list, but what kind?
    first = choices[0]
    # list of strings
    if type(first) == str:
        return Odict(choices, choices)
    # list of 2-element string lists
    elif type(first) == list and len(first) == 2:
        choices, values = zip(*choices)
        return Odict(choices, values)
    else:
        raise TypeError("choices lists must either be"\
            "['a', 'b', 'c'] or [['a', 'A'], ['b', 'B']] style.")



#! /usr/bin/env python
# opts.py

# WARNING: This module is a hack in progress

__all__ = [\
    'Option',
    'Usage',
    'parse',
    'validate']

### --------------------------------------------------------------------
import textwrap
from libtovid.odict import Odict
from libtovid.utils import trim

class Option (object):
    """A command-line-style option, expected argument formatting, default value,
    and notes on usage and purpose.
    
    For example:
        
        debug_opt = Option(
            'debug',
            'none|some|all',
            'some',
            "Amount of debugging information to display"
           )

    This defines a 'debug' option, along with a human-readable string showing
    expected argument formatting, a default value, and a string documenting the
    option's purpose and/or usage information."""

    def __init__(self, name, argformat='', default=None,
                 doc='Undocumented option', alias=None,
                 required=False):
        """Create a new option definition with the given attributes.
        
            name:      Option name
            argformat: String describing format of expected arguments
                       (see is_valid() for examples)
            default:   Default value, if any
            doc:       Manual-page-style documentation of the option
            alias:     An ('option', 'value') equivalent for this option.

        """
        self.name = name
        self.argformat = argformat
        self.default = default
        self.doc = trim(doc)
        self.alias = alias
        self.required = required
        # If an alias was provided, generate documentation.
        # i.e., alias=('tvsys', 'ntsc') means this option is the same as
        # giving the 'tvsys' option with 'ntsc' as the argument.
        if self.alias:
            self.doc = 'Same as -%s %s.' % alias

    def num_args(self):
        """Return the number of arguments expected by this option, or -1 if
        unlimited."""
        # Flag alias for another option
        if self.alias:
            return 0
        elif isinstance(self.default, bool):
            # Boolean: no argument
            return 0
        elif isinstance(self.default, list):
            # List: unlimited arguments
            return -1
        else:
            # Unary: one argument
            return 1
    
    def __str__(self):
        """Return a string containing "usage notes" for this option."""
        if self.alias:
            usage = "-%s: Same as '-%s %s'\n" % \
                  (self.name, self.alias[0], self.alias[1])
        else:
            usage = "-%s %s " % (self.name, self.argformat)
            if self.required:
                usage += "(REQUIRED)\n"
            else:
                usage += "(default: %s)\n" % self.default
            for line in textwrap.wrap(self.doc, 60):
                usage += '    ' + line + '\n'
        return usage

### --------------------------------------------------------------------

class Usage (object):
    """Command-line usage definition."""
    def __init__(self, usage_string='program [options]', *options):
        """Define usage of a command-line program.
        
            usage_string: Command-line syntax and required options
            *options:     List of allowed Options

        Examples:

        usage = Usage('pytovid [options] -in FILENAME -out NAME',
            Option('in', 'FILENAME', None,
                "Input video file, in any format."),
            Option('out', 'NAME', None,
                "Output prefix or name."),
            Option('format', 'vcd|svcd|dvd|half-dvd|dvd-vcd', 'dvd',
                "Make video compliant with the specified format")
        )
        
        print usage
        print usage.options['format']
        
        """
        self.usage_string = usage_string
        # Odict of options indexed by name
        names = [opt.name for opt in options]
        self.options = Odict(names, options)

    def __str__(self):
        """Return string-formatted usage notes."""
        result = "Usage: %s\n" % self.usage_string
        result += "Allowed options:\n"
        option_list = ['-' + opt.name for opt in self.options.values()]
        result += ', '.join(option_list)
        return result

### --------------------------------------------------------------------
from copy import copy

def parse(args):
    """Parse a list of arguments and return a dictionary of found options.

        args:    List of command-line arguments (such as from sys.argv)
    
    The argument list is interpreted in this way:
    
        - If it begins with '-', it's an option
        - Arguments that precede the first option are ignored
        - Anything following an option is an argument to that option
        - If there is no argument to an option, it's a flag (True if present)
        - If there's one argument to an option, it's a single string value
        - If there are multiple arguments to an option, it's a list of strings

    """
    args = copy(args)
    options = {}
    current = None
    while len(args) > 0:
        arg = args.pop(0)
        # New option
        if arg.startswith('-'):
            # Assume it's a flag unless we see arguments later
            current = arg.lstrip('-')
            options[current] = True
            
        # Argument to current option
        elif current:
            # Was a flag, now has a single value
            if options[current] == True:
                options[current] = arg
            # Was a single value, now a list
            elif type(options[current]) != list:
                options[current] = [options[current], arg]
            # Was a list, so append new value
            else:
                options[current].append(arg)
    return options


### --------------------------------------------------------------------
import re

def validate(option, arg):
    """Check whether an argument is valid for a given option.
    
        option: An Option to validate
        arg:    Candidate argument

    Expected/allowed values are inferred from the argformat string.
    
    argformat patterns to consider:
        opta|optb|optc      # Either opta, optb, or optc
        [100-999]           # Any integer between 100 and 999
        NUM:NUM             # Any two integers separated by a colon
        VAR                 # A single string (VAR can be any all-caps word)
        VAR [, VAR]         # List of strings
        (empty)             # Boolean; no argument required

    All very experimental...
    """
    # TODO: Eliminate hackery; find a more robust way of doing this.
    # Also, pretty inefficient, since regexp matching is done every
    # time this function is called.

    # Empty argformat means boolean/no argument expected
    if option.argformat == '':
        if arg in [True, False, '']:
            return True
        else:
            return False

    # Any alphanumeric word (any string)
    if re.compile('^\w+$').match(option.argformat):
        if arg.__class__ == str:
            return True
        else:
            return False

    # Square-bracketed values are ranges, i.e. [1-99]
    match = re.compile('^\[\d+-\d+\]$').match(option.argformat)
    if match:
        # Get min/max by stripping and splitting argformat
        limits = re.split('-', match.group().strip('[]'))
        option.min = int(limits[0])
        option.max = int(limits[1])
        if int(arg) >= option.min and int(arg) <= option.max:
            return True
        else:
            return False

    # Values separated by | are mutually-exclusive
    if re.compile('[-\w]+\|[-\w]+').match(option.argformat):
        if arg in option.argformat.split('|'):
            return True
        else:
            return False

    # For now, accept any unknown cases
    return True


import sys
if __name__ == '__main__':
    print "Option-parsing demo"

    print "You passed the following arguments:"
    print sys.argv[1:]

    print "Parsing..."
    options = parse(sys.argv[1:])

    print "Found the following options:"
    print options
    



"""Identifiers for something or another.
"""

# Identifier for command-panel timer
ID_CMD_TIMER = 101
# Identifiers for output-format radio buttons
ID_FMT_VCD = 0
ID_FMT_SVCD = 1
ID_FMT_DVD = 2
ID_FMT_HALFDVD = 3
ID_FMT_DVDVCD = 4
# Identifiers for tv systems
ID_TVSYS_NTSC = 0
ID_TVSYS_NTSCFILM = 1
ID_TVSYS_PAL = 2
# Identifiers for aspect ratios
ID_ASPECT_FULL = 0
ID_ASPECT_WIDE = 1
ID_ASPECT_PANA = 2
# Identifiers for alignments
ID_ALIGN_LEFT = 0
ID_ALIGN_CENTER = 1
ID_ALIGN_RIGHT = 2
ID_ALIGN_WEST = 3
ID_ALIGN_MIDDLE = 4
ID_ALIGN_EAST = 5
ID_ALIGN_SOUTHWEST = 6
ID_ALIGN_SOUTH = 7
ID_ALIGN_SOUTHEAST = 8
# Identifiers for test fill types
ID_TEXTFILL_COLOR = 0
ID_TEXTFILL_FRACTAL = 1
ID_TEXTFILL_GRADIENT = 2
ID_TEXTFILL_PATTERN = 3
# Identifiers for notebook tabs
ID_NB_VIDEO = 0
ID_NB_MENU = 1
# Identifiers for disc tree element types
ID_DISC = 0
ID_MENU = 1
ID_VIDEO = 2
ID_SLIDE = 3
ID_GROUP = 4
# Identifiers for menu items
ID_MENU_FILE_PREFS = 101
ID_MENU_FILE_EXIT = 102
ID_MENU_FILE_SAVE = 103
ID_MENU_FILE_OPEN = 104
ID_MENU_FILE_NEW = 105
ID_MENU_VIEW_SHOWGUIDE = 201
ID_MENU_VIEW_SHOWTOOLTIPS = 202
ID_MENU_VIEW_ADVANCED = 203
ID_MENU_HELP_ABOUT = 301
ID_MENU_LANG = 400
ID_MENU_LANG_EN = 401
ID_MENU_LANG_DE = 402
# Identifiers for tasks, for the guide and help system
ID_TASK_GETTING_STARTED = 1001
ID_TASK_MENU_ADDED = 1002
ID_TASK_VIDEO_ADDED = 1003
ID_TASK_DISC_SELECTED= 1004
ID_TASK_VIDEO_SELECTED = 1005
ID_TASK_MENU_SELECTED = 1006
ID_TASK_PREP_ENCODING = 1007
ID_TASK_ENCODING_STARTED = 1008
ID_TASK_BURN_DISC = 1009
ID_TASK_GROUP_ADDED = 1010
ID_TASK_GROUP_SELECTED = 1011
# Identifiers for disc authoring
ID_BURN_VCD_XML = 1
ID_BURN_DVD_XML = 2
ID_BURN_FILE = 3


id_dict = {
    'tvsys': {
        ID_TVSYS_NTSC: "ntsc",
        ID_TVSYS_NTSCFILM: "ntscfilm",
        ID_TVSYS_PAL: "pal"
    },
    'format': {
        ID_FMT_DVD: "dvd",
        ID_FMT_HALFDVD: "half-dvd",
        ID_FMT_DVDVCD: "dvd-vcd",
        ID_FMT_SVCD: "svcd",
        ID_FMT_VCD: "vcd"
    },
    'aspect': {
        ID_ASPECT_FULL: "full",
        ID_ASPECT_WIDE: "wide",
        ID_ASPECT_PANA: "panavision"
    },
    'alignment': {
        ID_ALIGN_LEFT:      "northwest",
        ID_ALIGN_CENTER:    "north",
        ID_ALIGN_RIGHT:     "northeast",
        ID_ALIGN_WEST:      "west",
        ID_ALIGN_MIDDLE:    "middle",
        ID_ALIGN_EAST:      "east",
        ID_ALIGN_SOUTHWEST: "southwest",
        ID_ALIGN_SOUTH:     "south",
        ID_ALIGN_SOUTHEAST: "southeast"
    },
    'fillType': {
        ID_TEXTFILL_COLOR:    "Color",
        ID_TEXTFILL_FRACTAL:  "Fractal",
        ID_TEXTFILL_GRADIENT: "Gradient",
        ID_TEXTFILL_PATTERN:  "Pattern"
    }
}



#! /usr/bin/env python
# gui.py

"""Classes for creating and laying out GUI applications.
"""

__all__ = [
    'Panel',
    'HPanel',
    'VPanel',
    'Dropdowns',
    'Drawer',
    'Tabs',
    'Application',
    'GUI']

import sys
import Tkinter as tk

from widget import Widget
from control import Control, MissingOption
from libtovid.cli import Command

### --------------------------------------------------------------------
### GUI interface-definition API
### --------------------------------------------------------------------

class Panel (Widget):
    """A group of option controls in a rectangular frame.

    For example:

        Panel("General",
            Filename('bgaudio', "Background audio file"),
            Flag('submenus', "Create submenus"),
            Number('menu-length', "Length of menu (seconds)")
            )

    This creates a panel with three GUI widgets that control command-line
    options '-bgaudio', '-submenus', and '-menu-length'.
    """
    def __init__(self, title='', *contents):
        """Create a Panel to hold control widgets or sub-panels.
        
            title:    Title of panel (name shown in tab bar)
            contents: One or more Widgets (Controls, Panels, Drawers etc.)
        """
        Widget.__init__(self)
        if type(title) != str:
            raise TypeError("First argument to Panel must be a text label.")
        self.title = title
        self.contents = []
        for item in contents:
            if isinstance(item, Widget):
                self.contents.append(item)
            else:
                import traceback
                # Print helpful info from stack trace
                last_error = traceback.extract_stack()[0]
                filename, lineno, foo, code = last_error
                print "Error on line %(lineno)s of %(filename)s:" % vars()
                print "    " + code
                print "Panel '%s' may only contain Widget subclasses" \
                      " (Controls, Panels, Drawers etc.)" \
                      " got %s instead" % type(item)
                sys.exit(1)

    def draw(self, master, side='top'):
        """Draw Panel and its contents in the given master.
        """
        Widget.draw(self, master)
        # Get a labeled or unlabeled frame
        if self.title:
            frame = tk.LabelFrame(self, text=self.title)
            frame.pack(fill='both', expand=True)
        else:
            frame = self
        # Draw all contents in the frame
        for item in self.contents:
            item.draw(frame)
            item.pack(side=side, anchor='nw', fill='x', expand=True,
                  padx=4, pady=4)

    def get_args(self):
        """Return a list of all command-line options from contained widgets.
        Print error messages if any required options are missing.
        """
        args = []
        for item in self.contents:
            try:
                args += item.get_args()
            except MissingOption, missing:
                print "Missing a required option: " + missing.option
        return args

### --------------------------------------------------------------------

class HPanel (Panel):
    """A panel with widgets packed left-to-right"""
    def __init__(self, title='', *contents):
        Panel.__init__(self, title, *contents)

    def draw(self, master):
        Panel.draw(self, master, 'left')

### --------------------------------------------------------------------

class VPanel (Panel):
    """A panel with widgets packed top-to-bottom"""
    def __init__(self, title='', *contents):
        Panel.__init__(self, title, *contents)

    def draw(self, master):
        Panel.draw(self, master, 'top')

### --------------------------------------------------------------------
from support import ComboBox, ListVar
from control import Control
from libtovid.odict import Odict

class Dropdowns (Panel):
    """A Panel that uses dropdowns for selecting and setting options.
    
    Given a list of controls, the Dropdowns panel displays, initially,
    a single dropdown list. Each control is a choice in the list, and
    is shown as two columns, option and label (with help shown in a tooltip).
    
    Selecting a control causes that control to be displayed, so that it
    may be set to the desired value (along with a "remove" button to discard
    the control). The dropdown list is shifted downward, so another control
    and option may be set.
    """
    def __init__(self, title='', *contents):
        Panel.__init__(self, title, *contents)
        # Controls, indexed by option
        self.controls = Odict()
        for control in self.contents:
            if not isinstance(control, Control):
                raise TypeError("Can only add Controls to a Dropdown")
            self.controls[control.option] = control

    def draw(self, master):
        if self.title:
            tk.LabelFrame.__init__(self, master, text=self.title,
                                   padx=8, pady=8)
        else:
            tk.LabelFrame.__init__(self, master, bd=0, text='',
                                   padx=8, pady=8)
        self.choices = ListVar(items=self.controls.keys())
        self.chosen = tk.StringVar()
        self.chooser = ComboBox(self, self.choices, variable=self.chosen,
                                command=self.choose_new)
        self.chooser.pack(fill='both', expand=True)

    def choose_new(self, event=None):
        """Create and display the chosen control."""
        chosen = self.chosen.get()
        if chosen == '':
            return
        self.chooser.pack_forget()
        # Put control and remove button in a frame
        frame = tk.Frame(self)
        button = tk.Button(frame, text="X",
                           command=lambda:self.remove(chosen))
        button.pack(side='left')
        control = self.controls[chosen]
        control.draw(frame)
        control.pack(side='left', fill='x', expand=True)
        frame.pack(fill='x', expand=True)
        # Remove the chosen control/panel from the list of available ones
        self.choices.remove(chosen)
        self.chooser.pack()

    def remove(self, option):
        """Remove a given option's control from the interface."""
        frame = self.controls[option].master
        frame.pack_forget()
        frame.destroy()
        # Make this option available in the dropdown
        self.choices.append(option)

    def get_args(self):
        """Return a list of all command-line options from contained widgets.
        """
        args = []
        for control in self.contents:
            if control.active:
                args += control.get_args()
        return args


### --------------------------------------------------------------------

class Drawer (Widget):
    """Like a Panel, but may be hidden or "closed" like a drawer."""
    def __init__(self, title='', *contents):
        Widget.__init__(self)
        self.panel = Panel(title, *contents)
        self.visible = False
        
    def draw(self, master):
        Widget.draw(self, master)
        # Checkbutton
        button = tk.Button(self, text=self.panel.title,
                           command=self.show_hide)
        button.pack(anchor='nw', fill='x', expand=True)
        # Draw panel, but don't pack
        self.panel.draw(self)

    def show_hide(self):
        # Hide if showing
        if self.visible:
            self.panel.pack_forget()
            self.visible = False
        # Show if hidden
        else:
            self.panel.pack(anchor='nw', fill='both', expand=True)
            self.visible = True
    
    def get_args(self):
        return self.panel.get_args()

### --------------------------------------------------------------------

class Tabs (Widget):
    """A widget with tab buttons that switch between several panels.
    """
    def __init__(self, *panels, **kwargs):
        """Create tabs that switch between several Panels.
        """
        Widget.__init__(self)
        self.index = 0
        self.panels = []
        for panel in panels:
            if not isinstance(panel, Panel):
                raise TypeError("Tabs may only contain Panels")
            self.panels.append(panel)

    def add(self, panel):
        """Add the given Panel to the Tabs.
        """
        self.panels.append(panel)

    def draw(self, master, side='top'):
        """Draw the Tabs widget in the given master."""
        Widget.draw(self, master)
        self.selected = tk.IntVar()
        self.side = side
        # Tkinter configuration common to all tab buttons
        config = {
            'variable': self.selected,
            'command': self.change,
            'selectcolor': 'white',
            'relief': 'sunken',
            'offrelief': 'groove',
            'indicatoron': 0,
            'padx': 4, 'pady': 4
            }
        # Frame to hold tab buttons
        self.buttons = tk.Frame(self)
        # For tabs on left or right, pack tab buttons vertically
        if self.side in ['left', 'right']:
            button_side = 'top'
            bar_anchor = 'n'
            bar_fill = 'y'
        else:
            button_side = 'left'
            bar_anchor = 'w'
            bar_fill = 'x'
        # Tab buttons, numbered from 0
        for index, panel in enumerate(self.panels):
            button = tk.Radiobutton(self.buttons, text=panel.title,
                                    value=index, **config)
            button.pack(anchor='nw', side=button_side,
                        fill='both', expand=True)
            panel.draw(self)
        self.buttons.pack(anchor=bar_anchor, side=self.side,
                          fill=bar_fill)
        # Activate the first tab
        self.selected.set(0)
        self.change()

    def change(self):
        """Switch to the selected tab's frame.
        """
        # Unpack the existing panel
        self.panels[self.index].pack_forget()
        # Pack the newly-selected panel
        selected = self.selected.get()
        self.panels[selected].pack(side=self.side, fill='both', expand=True)
        # Remember this tab's index
        self.index = selected

    def get_args(self):
        """Return command-line arguments from each of the tab Panels.
        """
        args = []
        for panel in self.panels:
            args += panel.get_args()
        return args

### --------------------------------------------------------------------
import tkMessageBox

class Application (Widget):
    """Graphical frontend for a command-line program
    """
    def __init__(self, program, panels=None, style=None):
        """Define a GUI application frontend for a command-line program.
        
            program: Command-line program that the GUI is a frontend for
            panels:  List of Panels (groups of widgets), containing controls
                     for the given program's options. Use [panel] to pass
                     a single panel. If there are multiple panels, a tabbed
                     application is created.

        After defining the Application, call run() to show/execute it.
        """
        Widget.__init__(self)
        self.program = program
        # TODO: Friendlier error-handling
        if not type(panels) == list or len(panels) == 0:
            raise TypeError("Application needs a list of Panels")
        self.panels = panels or []
        self.showing = False
        self.frame = None

    def draw(self, master):
        """Draw the Application in the given master.
        """
        Widget.draw(self, master)
        # Single-panel application
        if len(self.panels) == 1:
            panel = self.panels[0]
            panel.draw(self)
            panel.pack(anchor='n', fill='x', expand=True)
        # Multi-panel (tabbed) application
        else:
            tabs = Tabs(*self.panels)
            tabs.draw(self)
            tabs.pack(anchor='n', fill='x', expand=True)
        # "Run" button
        button = tk.Button(self, text="Run %s now" % self.program,
                           command=self.execute)
        button.pack(anchor='s', fill='x')

    def get_args(self):
        """Get a list of all command-line arguments from all panels.
        """
        if isinstance(self.panels, Panel):
            return self.panels.get_args()
        elif isinstance(self.panels, list):
            args = []
            for panel in self.panels:
                args += panel.get_args()
            return args

    def execute(self):
        """Run the program with all the supplied options.
        """
        args = self.get_args()
        command = Command(self.program, *args)

        print "Running command:", command
        # Verify with user
        if tkMessageBox.askyesno(message="Run %s now?" % self.program):
            # Kind of hackish...
            root = self.master.master
            root.withdraw()
            try:
                command.run()
            except KeyboardInterrupt:
                tkMessageBox.showerror(message="todisc was interrupted!")
            else:
                tkMessageBox.showinfo(message="todisc finished running!")
            root.deiconify()

### --------------------------------------------------------------------
from support import ConfigWindow, Style
import os
DEFAULT_CONFIG = os.path.expanduser('~/.metagui/config')

class GUI (tk.Tk):
    """GUI with one or more Applications
    """
    def __init__(self, title, applications, width=600, height=600,
                 inifile=None):
        """Create a GUI for the given applications.
        
            title:        Text shown in the title bar
            applications: List of Applications to included in the GUI
            width:        Initial width of GUI window in pixels
            height:       Initial height of GUI window in pixels
            inifile:      Name of an .ini-formatted file with GUI configuration
        """
        tk.Tk.__init__(self)
        self.geometry("%dx%d" % (width, height))
        self.title(title)
        # TODO: Friendlier error-handling
        if not type(applications) == list or len(applications) == 0:
            raise TypeError("GUI needs a list of Applications")
        self.apps = applications
        self.width = width
        self.height = height
        self.inifile = inifile or DEFAULT_CONFIG
        self.style = Style()
        if os.path.exists(self.inifile):
            print "Loading style from config file", self.inifile
            self.style.load(self.inifile)
        else:
            print "Creating config file", self.inifile
            self.style.save(self.inifile)
        # Show hidden file option in file dialogs
        self.tk.call('namespace', 'import', '::tk::dialog::file::')
        self.tk.call('set', '::tk::dialog::file::showHiddenBtn',  '1')
        self.tk.call('set', '::tk::dialog::file::showHiddenVar',  '0')

    def run(self):
        """Run the GUI"""
        self.draw()
        self.draw_menu(self)
        # Enter the main event handler
        self.mainloop()
        # TODO: Interrupt handling

    def draw(self):
        """Draw widgets."""
        self.style.apply(self)
        self.frame = tk.Frame(self, width=self.width, height=self.height)
        self.frame.pack(fill='both', expand=True)
        self.resizable(width=True, height=True)
        # Single-application GUI
        if len(self.apps) == 1:
            app = self.apps[0]
            app.draw(self.frame)
            app.pack(anchor='n', fill='both', expand=True)
        # Multi-application (tabbed) GUI
        else:
            tabs = Tabs(self.frame, 'top')
            for app in self.apps:
                app.draw(tabs)
                tabs.add(app.program, app)
            tabs.draw()
            tabs.pack(anchor='n', fill='both', expand=True)
        #self.frame.pack_propagate(False)
    
    def redraw(self):
        self.frame.destroy()
        self.draw()

    def draw_menu(self, window):
        """Draw a menu bar in the given top-level window.
        """
        # Create and add the menu bar
        menubar = tk.Menu(window)
        window.config(menu=menubar)
        # File menu
        filemenu = tk.Menu(menubar, tearoff=False)
        filemenu.add_command(label="Config", command=self.show_config)
        filemenu.add_separator()
        filemenu.add_command(label="Exit", command=self.quit)
        menubar.add_cascade(label="File", menu=filemenu)

    def show_config(self):
        """Open the GUI configuration dialog."""
        config = ConfigWindow(self, self.style)
        if config.result:
            self.style = config.result
            print "Saving configuration to", self.inifile
            self.style.save(self.inifile)
            self.style.apply(self)
            self.redraw()

### --------------------------------------------------------------------




#! /usr/bin/env python
# cli.py

"""This module provides an interface for running command-line applications.
Two primary classes are provided:

    Command:  For constructing and executing command-line commands
    Pipe:     For piping commands together
    
Commands are constructed by specifying a program to run, and each separate
argument to pass to that program. Arguments are used in a platform-independent
way, so there is no need to escape shell-specific characters or do any quoting
of arguments.

Commands may be executed in the foreground or background; they can print their
output on standard output, or capture it in a string variable.

For example:

    >>> echo = Command('echo', "Hello world")
    >>> echo.run()                                # doctest: +SKIP
    Hello world

Commands may be connected together with pipes:

    >>> sed = Command('sed', 's/world/nurse/')
    >>> pipe = Pipe(echo, sed)
    >>> pipe.run()                                # doctest: +SKIP
    Hello nurse

Command output may be captured and retrieved later with get_output():

    >>> echo.run(capture=True)
    >>> echo.get_output()
    'Hello world\\n'

"""
# Note: Some of the run() tests above will fail doctest.testmod(), since output
# from Command subprocesses is not seen as real output by doctest. The current
# workaround is to use the "doctest: +SKIP" directive (new in python 2.5).
# For other directives see http://www.python.org/doc/lib/doctest-options.html

__all__ = [\
    'Command',
    'Pipe']

import os
import sys
import doctest
import signal
from subprocess import Popen, PIPE

class Command:
    """A command-line statement, consisting of a program and its arguments,
    with support for various modes of execution.
    """
    def __init__(self, program, *args):
        """Create a Command to run a program with the given arguments.
        
            program: A string containing the name of a program to execute
            args:    Individual arguments to supply the command with
        
        For example:
        
            >>> cmd = Command('echo', 'Hello world')
            
        """
        self.program = program
        self.args = []
        for arg in args:
            self.add(arg)
        self.proc = None
        self.output = ''

    def add(self, *args):
        """Append one or more arguments to the command. Each argument passed
        to this function is converted to string form, and treated as a single
        argument in the command. No special quoting or escaping of argument
        contents is necessary.
        """
        for arg in args:
            self.args.append(str(arg))
    
    def run(self, capture=False, background=False):
        """Run the command and capture or display output.
        
            capture:    False to show command output/errors on stdout,
                        True to capture output/errors for retrieval
                        by get_output() and get_error()
            background: False to wait for command to finish running,
                        True to run process in the background
        
        By default, this function displays all command output, and waits
        for the program to finish running, which is usually what you'd want.
        Capture output if you don't want it printed immediately (and call
        get_output() later to retrieve it).

        This function does not allow special stream redirection. For that,
        use run_redir().
        """
        print "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
        print "Running command:"
        print str(self)
        print "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"

        if capture:
            self.run_redir(None, PIPE, stderr=PIPE)
        else:
            self.run_redir(None, None, stderr=None)
        if not background:
            self.wait()

    def run_redir(self, stdin=None, stdout=None, stderr=None):
        """Execute the command using the given stream redirections.
        
            stdin:  Filename or File object to read input from
            stdout: Filename or File object to write output to
            stderr: Filename or File object to write errors to
        
        Use None for regular system stdin/stdout/stderr (default behavior).
        That is, if stdout=None, the command's standard output is printed.
        
        This function is used internally by run(); if you need to do stream
        redirection (ex. `spumux < menu.mpg > menu_subs.mpg`), use this
        function instead of run(), and call wait() afterwards if needed.
        """
        self.output = ''
        # Open files if string filenames were provided
        if type(stdin) == str:
            stdin = open(stdin, 'r')
        if type(stdout) == str:
            stdout = open(stdout, 'w')
        if type(stderr) == str:
            stderr = open(stderr, 'w')
        # Run the subprocess
        self.proc = Popen([self.program] + self.args,
                          stdin=stdin, stdout=stdout, stderr=stderr)

    def wait(self):
        """Wait for the command to finish running. If a KeyboardInterrupt
        occurs (user pressed Ctrl-C), kill the subprocess (and re-raise the
        KeyboardInterrupt exception).
        """
        if not isinstance(self.proc, Popen):
            print "**** Can't wait(): Command is not running"
            return
        try:
            self.proc.wait()
        except KeyboardInterrupt:
            self.kill()
            raise KeyboardInterrupt

    def kill(self):
        """Abort!"""
        os.kill(self.proc.pid, signal.SIGTERM)

    def done(self):
        """Return True if the command is finished running; False otherwise.
        Only useful if the command is run in the background.
        """
        return self.proc.poll() != None

    def get_output(self):
        """Wait for the command to finish running, and return a string
        containing the command's output. If this command is piped into another,
        return that command's output instead. Returns an empty string if the
        command has not been run yet.
        """
        if self.output is '' and isinstance(self.proc, Popen):
            self.output = self.proc.communicate()[0]
        return self.output

    def get_error(self):
        """Wait for the command to finish running, and return a string
        containing the command's standard error output. Returns an empty
        string if the command has not been run yet.
        """
        if self.error is '' and isinstance(self.proc, Popen):
            self.error = self.proc.communicate()[1]
        return self.error

    get_errors = get_error

    def __str__(self):
        """Return a string representation of the Command, as it would look if
        run in a command-line shell.
        """
        ret = self.program
        for arg in self.args:
            ret += " %s" % _enc_arg(arg)
        return ret


class Pipe:
    """A series of Commands, each having its output piped into the next.
    """
    def __init__(self, *commands):
        """Create a new Pipe containing all the given Commands."""
        self.commands = []
        for cmd in commands:
            self.add(cmd)
        self.proc = None
    
    def add(self, *commands):
        """Append the given commands to the end of the pipeline."""
        for cmd in commands:
            self.commands.append(cmd)

    def run(self, capture=False, background=False):
        """Run all Commands in the pipeline, doing appropriate stream
        redirection for piping.
        
            capture:    False to show pipeline output on stdout,
                        True to capture output for retrieval by get_output()
        
        """
        self.output = ''
        prev_stdout = None
        # Run each command, piping to the next
        for cmd in self.commands:
            # If this is not the last command, pipe into the next one
            if cmd != self.commands[-1]:
                cmd.run_redir(prev_stdout, PIPE)
                prev_stdout = cmd.proc.stdout
            # Last command in pipeline; direct output appropriately
            else:
                if capture:
                    cmd.run_redir(prev_stdout, PIPE)
                else:
                    cmd.run_redir(prev_stdout, None)
        # Wait for last command to finish?
        if not background:
            cmd.wait()

    def get_output(self):
        """Wait for the pipeline to finish executing, and return a string
        containing the output from the last command in the pipeline.
        """
        return self.commands[-1].get_output()

    def __str__(self):
        """Return a string representation of the Pipe.
        """
        commands = [str(cmd) for cmd in self.commands]
        return ' | '.join(commands)


def _enc_arg(arg):
    """Quote an argument for proper handling of special shell characters.
    Don't quote unless necessary. For example:

        >>> print _enc_arg("spam")
        spam
        >>> print _enc_arg("spam & eggs")
        'spam & eggs'
        >>> print _enc_arg("['&&']")
        '['\\''&&'\\'']'
    
    This is used internally by Command; you'd only need this if you're running
    shell programs without using the Command class.
    """
    arg = str(arg)
    # At the first sign of any special character in the argument,
    # single-quote the whole thing and return it (escaping ' itself)
    for char in ' #"\'\\&|<>()[]!?*':
        if char in arg:
            return "'%s'" % arg.replace("'", "'\\''")
    # No special characters found; use literal string
    return arg


if __name__ == '__main__':
    doctest.testmod(verbose=True)



#! /usr/bin/env python
# widget.py

__all__ = ['Widget']

import Tkinter as tk

### --------------------------------------------------------------------
class Widget (tk.Frame):
    """Generic metagui widget, suitable for Controls, Panels, etc."""
    def __init__(self):
        self.active = False
    
    def draw(self, master):
        tk.Frame.__init__(self, master)
        self.active = True

    def destroy(self):
        tk.Frame.destroy(self)
        self.active = False

    def enable(self, enabled=True):
        """Enable or disable all sub-widgets."""
        if enabled:
            newstate = 'normal'
        else:
            newstate = 'disabled'
        for widget in self.children.values():
            # Some widgets don't support state changes
            if 'state' in widget.config():
                widget.config(state=newstate)

    def disable(self):
        """Disable all sub-widgets."""
        self.enable(False)



#! /usr/bin/env python
# control.py

"""Control widget classes.

This module defines a Control class and several derivatives. A Control is a
special-purpose GUI widget for setting a value such as a number or filename.

"""

__all__ = [
    'Control',
    # Control subclasses
    'Choice',
    'Color',
    'Filename',
    'Flag',
    'FlagGroup',
    'Font',
    'Text',
    'List',
    'Number',
    'FileList',
    'TextList']

import Tkinter as tk
from widget import Widget
from support import ListVar
import support
import os

### --------------------------------------------------------------------
class MissingOption (Exception):
    def __init__(self, option):
        self.option = option

### --------------------------------------------------------------------
# Map python types to Tkinter variable types
_vartypes = {
    str: tk.StringVar,
    bool: tk.BooleanVar,
    int: tk.IntVar,
    float: tk.DoubleVar,
    list: ListVar}

### --------------------------------------------------------------------
from tooltip import ToolTip

class Control (Widget):
    """A widget that controls the value of a command-line option.

    A Control is a specialized GUI widget that controls a command-line option
    via a local variable, accessed via get() and set() methods.
    
    Control subclasses may have any number of sub-widgets such as labels,
    buttons or entry boxes; one of the sub-widgets should be linked to the
    controlled variable via an option like:
    
        entry = Entry(self, textvariable=self.variable)
    
    See the Control subclasses below for examples of how self.variable,
    get() and set() are used.
    """
    def __init__(self,
                 vartype=str,
                 label='',
                 option='',
                 default='',
                 help='',
                 **kwargs):
        """Create a Control for an option.

            vartype:  Type of stored variable (str, bool, int, float, list)
            label:    Label shown in the GUI for the Control
            option:   Command-line option associated with this Control, or
                      '' to create a positional argument
            default:  Default value for the Control
            help:     Help text to show in a tooltip
            **kwargs: Keyword arguments of the form key1=arg1, key2=arg2
        
        Keyword arguments allowed:
        
            pull=Control(...):  Mirror value from another Control
            required=True:      Required option, must be set or run will fail
            filter=function:    Text-filtering function for pulled values
            toggles=True:       May be toggled on and off
        """
        Widget.__init__(self)
        self.vartype = vartype
        self.variable = None
        self.label = label
        self.option = option
        self.default = default or vartype()
        self.help = help
        self.kwargs = kwargs

        # TODO: Find a way to condense/separate keyword handling
        # Controls a mandatory option?
        self.required = False
        if 'required' in kwargs:
            self.required = bool(kwargs['required'])
        # Has an enable/disable toggle button
        self.toggles = False
        if 'toggles' in self.kwargs:
            self.toggles = bool(self.kwargs['toggles'])
        # List of Controls to copy updated values to
        self.copies = []
        if 'pull' in self.kwargs:
            if not isinstance(self.kwargs['pull'], Control):
                raise TypeError("Can only pull values from a Control.")
            pull_from = self.kwargs['pull']
            pull_from.copy_to(self)
        # Filter expression when pulling from another Control
        self.filter = None
        if 'filter' in self.kwargs:
            if not callable(self.kwargs['filter']):
                raise TypeError("Pull filter must be a function.")
            self.filter = self.kwargs['filter']

    def copy_to(self, control):
        """Update another control whenever this control's value changes.
        """
        if not isinstance(control, Control):
            raise TypeError("Can only copy values to a Control.")
        self.copies.append(control)

    def draw(self, master):
        """Draw the control widgets in the given master.
        
        Override this method in derived classes, and call the base
        class draw() method:
        
            Control.draw(self, master)
        
        """
        Widget.draw(self, master)
        # Create tk.Variable to store Control's value
        if self.vartype in _vartypes:
            self.variable = _vartypes[self.vartype]()
        else:
            self.variable = tk.Variable()
        # Set default value
        if self.default:
            self.variable.set(self.default)
        # Draw tooltip
        if self.help != '':
            self.tooltip = ToolTip(self, text=self.help, delay=1000)
        # Draw enabler checkbox
        if self.toggles:
            self.enabled = tk.BooleanVar()
            self.check = tk.Checkbutton(self, text='',
                                        variable=self.enabled,
                                        command=self.enabler)
            self.check.pack(side='left')

    def post(self):
        """Post-draw initialization.
        """
        if self.toggles:
            self.enabler()

    def enabler(self):
        """Enable or disable the Control when the checkbox is toggled.
        """
        if self.enabled.get():
            self.enable()
        else:
            self.disable()
            self.check.config(state='normal')

    def get(self):
        """Return the value of the Control's variable."""
        # self.variable isn't set until draw() is called
        if not self.variable:
            raise "Must call draw() before get()"
        return self.variable.get()

    def set(self, value):
        """Set the Control's variable to the given value."""
        # self.variable isn't set until draw() is called
        if not self.variable:
            raise "Must call draw() before set()"
        self.variable.set(value)
        # Update controls that are copying this control's value
        for control in self.copies:
            control.variable.set(value)

    def get_args(self):
        """Return a list of arguments for passing this command-line option.
        draw() must be called before this function.
        """
        # TODO: Raise exception if draw() hasn't been called
        args = []
        value = self.get()

        # For toggles, return 
        if self.toggles:
            if not self.enabled.get():
                return []
        # Skip if unmodified or empty
        elif value == self.default or value == []:
            # ...unless it's required
            if self.required:
                raise MissingOption(self.option)
            else:
                return []

        # '-option'
        if self.option != '':
            args.append(self.option)
        # List of arguments
        if type(value) == list:
            args.extend(value)
        # Single argument
        else:
            args.append(value)
        return args

    def __repr__(self):
        """Return a Python code representation of this Control."""
        # Get derived class name
        control = str(self.__class__).split('.')[-1]
        return "%s('%s', '%s')" % (control, self.option, self.label)

### --------------------------------------------------------------------
### Control subclasses
### --------------------------------------------------------------------

class Flag (Control):
    """A widget for controlling a yes/no value."""
    def __init__(self,
                 label="Flag",
                 option='',
                 default=False,
                 help='',
                 **kwargs):
        """Create a Flag widget with the given label and default value.

            label:    Text label for the flag
            option:   Command-line flag passed
            default:  Default value (True or False)
            help:     Help text to show in a tooltip
        """
        Control.__init__(self, bool, label, option, default, help, **kwargs)
        # Enable an associated control when this Flag is True
        self.enables = None
        if 'enables' in kwargs:
            self.enables = kwargs['enables']
            if not isinstance(self.enables, Widget):
                raise "A Flag can only enable a Widget (Control or Panel)"

    def draw(self, master):
        """Draw control widgets in the given master."""
        Control.draw(self, master)
        self.check = tk.Checkbutton(self, text=self.label,
                                    variable=self.variable,
                                    command=self.enabler)
        self.check.pack(side='left')
        # Draw any controls enabled by this one
        if self.enables:
            self.enables.draw(self)
            self.enables.pack(side='left', fill='x', expand=True)
            # Disable if False
            if not self.default:
                self.enables.disable()
        Control.post(self)

    def enabler(self):
        """Enable/disable a Control based on the value of the Flag."""
        if not self.enables:
            return
        if self.get():
            self.enables.enable()
        else:
            self.enables.disable()

    def get_args(self):
        """Return a list of arguments for passing this command-line option.
        draw() must be called before this function.
        """
        args = []
        if self.get() == True:
            if self.option:
                args.append(self.option)
            if self.enables:
                args.extend(self.enables.get_args())
        return args

### --------------------------------------------------------------------

class FlagGroup (Control):
    """A wrapper widget for grouping Flag controls, and allowing
    mutually-exclusive flags.
    """
    def __init__(self,
                 label='',
                 state='normal',
                 *flags,
                 **kwargs):
        """Create a FlagGroup with the given label and state.
        
            label:    Label for the group
            state:    'normal' for regular Flags, 'exclusive' for
                      mutually-exclusive Flags
            *flags:   All additional arguments are Flag controls
        """
        Control.__init__(self, str, '', label, '', '')
        self.flags = flags
        self.state = state
        self.label = label
    
    def draw(self, master):
        """Draw Flag controls in the given master."""
        Control.draw(self, master)
        frame = tk.LabelFrame(self, text=self.label)
        frame.pack(fill='x', expand=True)
        for flag in self.flags:
            flag.draw(frame)
            flag.check.bind('<Button-1>', self.select)
            flag.pack(anchor='nw', side='top', fill='x', expand=True)
        Control.post(self)

    def select(self, event):
        """Event handler called when a Flag is selected."""
        # For normal flags, nothing to do
        if self.state != 'exclusive':
            return
        # For exclusive flags, clear all but the clicked Flag
        for flag in self.flags:
            if flag.check != event.widget:
                flag.set(False)
            flag.enabler()

    def get_args(self):
        """Return a list of arguments for setting the relevant flag(s)."""
        args = []
        for flag in self.flags:
            if flag.option != 'none':
                args.extend(flag.get_args())
        return args

### --------------------------------------------------------------------
from libtovid.odict import Odict, convert_list
from support import ComboBox

class Choice (Control):
    """A widget for choosing one of several options.
    """
    def __init__(self,
                 label="Choices",
                 option='',
                 default=None,
                 help='',
                 choices='A|B',
                 style='radio',
                 packside='left',
                 **kwargs):
        """Initialize Choice widget with the given label and list of choices.

            label:    Text label for the choices
            option:   Command-line option to set
            default:  Default choice, or None to use first choice in list
            help:     Help text to show in a tooltip
            choices:  Available choices, in string form: 'one|two|three'
                      or list form: ['one', 'two', 'three'], or as a
                      list-of-lists: [['a', "Use A"], ['b', "Use B"], ..].
                      A dictionary is also allowed, as long as you don't
                      care about preserving choice order.
            style:    'radio' for radiobuttons, 'dropdown' for a drop-down list
        """
        self.choices = convert_list(choices)
        Control.__init__(self, str, label, option,
                         default or self.choices.values()[0],
                         help, **kwargs)
        if style not in ['radio', 'dropdown']:
            raise ValueError("Choice style must be 'radio' or 'dropdown'")
        self.style = style
        self.packside = packside

    def draw(self, master):
        """Draw control widgets in the given master."""
        Control.draw(self, master)
        if self.style == 'radio':
            frame = tk.LabelFrame(self, text=self.label)
            frame.pack(anchor='nw', fill='x')
            self.rb = {}
            for choice, label in self.choices.items():
                self.rb[choice] = tk.Radiobutton(frame,
                    text=label, value=choice, variable=self.variable)
                self.rb[choice].pack(anchor='nw', side=self.packside)
        else: # dropdown/combobox
            tk.Label(self, text=self.label).pack(side='left')
            self.combo = ComboBox(self, self.choices.keys(),
                                  variable=self.variable)
            self.combo.pack(side='left')
        Control.post(self)

### --------------------------------------------------------------------

class Number (Control):
    """A widget for choosing or entering a number"""
    def __init__(self,
                 label="Number",
                 option='',
                 default=None,
                 help='',
                 min=1,
                 max=10,
                 style='spin',
                 units='',
                 **kwargs):
        """Create a number-setting widget.
        
            label:    Text label describing the meaning of the number
            option:   Command-line option to set
            default:  Default value, or None to use minimum
            help:     Help text to show in a tooltip
            min, max: Range of allowable numbers (inclusive)
            style:    'spin' for a spinbox, or 'scale' for a slider
            units:    Units of measurement (ex. "kbits/sec"), used as a label
        """
        if default is None:
            default = min
        Control.__init__(self, int, label, option, default,
                         help, **kwargs)
        self.min = min
        self.max = max
        self.style = style
        self.units = units

    def draw(self, master):
        """Draw control widgets in the given master."""
        Control.draw(self, master)
        tk.Label(self, name='label', text=self.label).pack(side='left')
        if self.style == 'spin':
            self.number = tk.Spinbox(self, from_=self.min, to=self.max,
                                     width=4, textvariable=self.variable)
            self.number.pack(side='left')
            tk.Label(self, name='units', text=self.units).pack(side='left')

        else: # 'scale'
            tk.Label(self, name='units', text=self.units).pack(side='left')
            self.number = tk.Scale(self, from_=self.min, to=self.max,
                                   tickinterval=(self.max - self.min),
                                   variable=self.variable, orient='horizontal')
            self.number.pack(side='left', fill='x', expand=True)
        Control.post(self)

    def enable(self, enabled=True):
        """Enable or disable all sub-widgets."""
        # Overridden to make Scale widget look disabled
        Control.enable(self, enabled)
        if self.style == 'scale':
            if enabled:
                self.number['fg'] = 'black'
                self.number['troughcolor'] = 'white'
            else:
                self.number['fg'] = '#A3A3A3'
                self.number['troughcolor'] = '#D9D9D9'
                

### --------------------------------------------------------------------

class Text (Control):
    """A widget for entering a line of text"""
    def __init__(self,
                 label="Text",
                 option='',
                 default='',
                 help='',
                 **kwargs):
        """
            label:    Label for the text
            option:   Command-line option to set
            default:  Default value of text widget
            help:     Help text to show in a tooltip
        """
        Control.__init__(self, str, label, option, default, help, **kwargs)

    def draw(self, master):
        """Draw control widgets in the given master."""
        Control.draw(self, master)
        tk.Label(self, text=self.label, justify='left').pack(side='left')
        self.entry = tk.Entry(self, textvariable=self.variable)
        self.entry.pack(side='left', fill='x', expand=True)
        Control.post(self)

### --------------------------------------------------------------------
import shlex

class List (Text):
    """A widget for entering a space-separated list of text items"""
    def __init__(self,
                 label="List",
                 option='',
                 default='',
                 help='',
                 **kwargs):
        Text.__init__(self, label, option, default, help, **kwargs)

    def draw(self, master):
        """Draw control widgets in the given master."""
        Text.draw(self, master)
        Text.post(self)

    def get(self):
        """Split text into a list at whitespace boundaries."""
        text = Text.get(self)
        return shlex.split(text)

    def set(self, listvalue):
        """Set a value to a list, joined with spaces."""
        text = ' '.join(listvalue)
        Text.set(self, text)
    

### --------------------------------------------------------------------
from tkFileDialog import asksaveasfilename, askopenfilename

class Filename (Control):
    """A widget for entering or browsing for a filename"""
    def __init__(self,
                 label='Filename',
                 option='',
                 default='',
                 help='',
                 action='load',
                 desc='Select a file to load',
                 **kwargs):
        """Create a Filename with label, text entry, and browse button.
        
            label:   Text of label next to file entry box
            option:  Command-line option to set
            default: Default filename
            help:    Help text to show in a tooltip
            action:  Do you intend to 'load' or 'save' this file?
            desc:    Brief description (shown in title bar of file
                     browser dialog)
        """
        Control.__init__(self, str, label, option, default, help, **kwargs)
        self.action = action
        self.desc = desc

    def draw(self, master):
        """Draw control widgets in the given master."""
        Control.draw(self, master)
        # Create and pack widgets
        tk.Label(self, text=self.label, justify='left').pack(side='left')
        self.entry = tk.Entry(self, textvariable=self.variable)
        self.button = tk.Button(self, text="Browse...", command=self.browse)
        self.entry.pack(side='left', fill='x', expand=True)
        self.button.pack(side='left')
        Control.post(self)

    def browse(self, event=None):
        """Event handler when browse button is pressed"""
        if self.action == 'save':
            filename = asksaveasfilename(parent=self, title=self.desc)
        else: # 'load'
            filename = askopenfilename(parent=self, title=self.desc)
        # Got a filename? Display it
        if filename:
            self.set(filename)

### --------------------------------------------------------------------
import tkColorChooser

class Color (Control):
    """A widget for choosing a color"""
    def __init__(self,
                 label="Color",
                 option='',
                 default='',
                 help='',
                 **kwargs):
        """Create a widget that opens a color-chooser dialog.
        
            label:   Text label describing the color to be selected
            option:  Command-line option to set
            default: Default color (named color or hexadecimal RGB)
            help:    Help text to show in a tooltip
        """
        Control.__init__(self, str, label, option, default, help, **kwargs)

    def draw(self, master):
        """Draw control widgets in the given master."""
        Control.draw(self, master)
        tk.Label(self, text=self.label).pack(side='left')
        self.button = tk.Button(self, text="None", command=self.change)
        self.button.pack(side='left')
        Control.post(self)

    def change(self):
        """Choose a color, and set the button's label and color to match."""
        rgb, color = tkColorChooser.askcolor(self.get())
        if color:
            self.set(color)
            self.button.config(text=color, foreground=color)
    
### --------------------------------------------------------------------

class Font (Control):
    """A font selector widget"""
    def __init__(self,
                 label='Font',
                 option='',
                 default='Helvetica',
                 help='',
                 **kwargs):
        """Create a widget that opens a font chooser dialog.
        
            label:   Text label for the font
            option:  Command-line option to set
            default: Default font
            help:    Help text to show in a tooltip
        """
        Control.__init__(self, str, label, option, default, help, **kwargs)

    def draw(self, master):
        """Draw control widgets in the given master."""
        Control.draw(self, master)
        tk.Label(self, text=self.label).pack(side='left')
        self.button = tk.Button(self, textvariable=self.variable,
                                command=self.choose)
        self.button.pack(side='left', padx=8)
        Control.post(self)

    def choose(self):
        """Open a font chooser to select a font."""
        chooser = support.FontChooser()
        if chooser.result:
            self.variable.set(chooser.result)

### --------------------------------------------------------------------
from support import DragList

class TextList (Control):
    """A widget for listing and editing several text strings"""
    def __init__(self,
                 label="Text list",
                 option='',
                 default=None,
                 help='',
                 **kwargs):
        Control.__init__(self, list, label, option, default, help, **kwargs)

    def draw(self, master):
        """Draw control widgets in the given master."""
        Control.draw(self, master)
        frame = tk.LabelFrame(self, text=self.label)
        frame.pack(fill='x', expand=True)
        self.selected = tk.StringVar()
        self.listbox = DragList(frame, choices=self.variable,
                                chosen=self.selected)
        self.listbox.pack(fill='x', expand=True)
        # TODO: Event handling to allow editing items
        self.editbox = tk.Entry(frame, width=30, textvariable=self.selected)
        self.editbox.bind('<Return>', self.setTitle)
        self.editbox.pack(fill='x', expand=True)
        Control.post(self)

    def setTitle(self, event):
        """Event handler when Enter is pressed after editing a title."""
        index = self.listbox.curindex
        self.variable[index] = self.selected.get()
        # TODO: Select next item in list and focus the editbox

### --------------------------------------------------------------------
from tkFileDialog import askopenfilenames

class FileList (Control):
    """A widget for listing several filenames"""
    def __init__(self,
                 label="File list",
                 option='',
                 default=None,
                 help='',
                 **kwargs):
        Control.__init__(self, list, label, option, default, help, **kwargs)

    def draw(self, master):
        """Draw control widgets in the given master."""
        Control.draw(self, master)
        frame = tk.LabelFrame(self, text=self.label)
        frame.pack(fill='x', expand=True)
        # List of files
        self.listbox = DragList(frame, choices=self.variable,
                                command=self.select)
        self.listbox.pack(fill='x', expand=True)
        # Add/remove buttons
        group = tk.Frame(frame)
        self.add = tk.Button(group, text="Add...", command=self.addFiles)
        self.remove = tk.Button(group, text="Remove", command=self.removeFiles)
        self.add.pack(side='left', fill='x', expand=True)
        self.remove.pack(side='left', fill='x', expand=True)
        group.pack(fill='x')
        Control.post(self)

    def select(self, event=None):
        """Event handler when a filename in the list is selected.
        """
        pass

    def addFiles(self):
        """Event handler to add files to the list"""
        files = askopenfilenames(parent=self, title='Add files')
        self.listbox.add(*files)
        for dest in self.copies:
            self.listbox.linked = dest.listbox
            dest.listbox.linked = self.listbox
            if dest.filter:
                titles = [dest.filter(file) for file in files]
                dest.listbox.add(*titles)
            else:
                dest.listbox.add(*files)

    def removeFiles(self):
        """Event handler to remove selected files from the list"""
        # TODO: Support multiple selection
        selected = self.listbox.curindex
        self.listbox.delete(selected)
        for control in self.copies:
            control.listbox.delete(selected)

### --------------------------------------------------------------------

# Exported control classes, indexed by name
CONTROLS = {
    'Choice': Choice,
    'Color': Color,
    'Filename': Filename,
    'Flag': Flag,
    'FlagGroup': FlagGroup,
    'Font': Font,
    'Text': Text,
    'List': List,
    'Number': Number,
    'FileList': FileList,
    'TextList': TextList}

### --------------------------------------------------------------------

# Demo
if __name__ == '__main__':
    root = tk.Tk()
    for name, control in CONTROLS.items():
        frame = tk.LabelFrame(root, text=name, padx=10, pady=10,
                          font=('Helvetica', 10, 'bold'))
        frame.pack(fill='both', expand=True)
        widget = control()
        widget.draw(frame)
        widget.pack(fill='both')
    root.mainloop()
    



#! /usr/bin/env python
# author.py

"""This module is for authoring a DVD or (S)VCD using dvdauthor or vcdimager.
"""

__all__ = [\
    'Disc',
    'Menu',
    'Video',
    'Group',
    'Titleset',
    'vcdimager_xml',
    'dvdauthor_xml']

from libtovid import xml

###
### Supporting classes
###

class Video:
    """A video title for inclusion on a video disc.
    """
    def __init__(self, filename, title=''):
        self.filename = filename
        self.title = title
        self.chapters = []


class Group:
    """A group title for inclusion on a video disc.
    """
    def __init__(self, filenames, title):
        self.filenames = filenames
        self.title = title


class Menu:
    """A menu for navigating the titles on a video disc.
    """
    def __init__(self, filename='', videos=None):
        """Create a menu linking to the given Videos."""
        self.filename = filename
        self.videos = videos or []

    def add(self, video):
        """Add a Video to the Menu."""
        self.videos.append(video)


class Titleset:
    """A group of videos, with an optional Menu.
    """
    def __init__(self, menu=None, videos=None):
        """Create a Titleset containing the given Videos.
        """
        self.menu = menu
        self.videos = videos or []

    def add(self, video):
        """Add a Video to the Titleset."""
        self.videos.append(video)


class Disc:
    """A video disc containing one or more Titlesets, and an optional
    top Menu for navigating to each Titleset.
    
    """
    def __init__(self, name='Untitled', format='dvd', tvsys='ntsc',
                 titlesets=None):
        """Create a Disc with the given properties.
        
            format:    'vcd', 'svcd', or 'dvd'
            tvsys:     'pal' or 'ntsc'
            title:     String containing the title of the disc
            titlesets: List of Titlesets
        """
        self.name = name
        self.format = format
        self.tvsys = tvsys
        self.topmenu = None
        self.titlesets = titlesets or []



###
### Internal functions
###

def _add_titleset(titleset, ts_id, segment_items, sequence_items, pbc):
    """Add titleset content to a vcdimager XML structure. This function is
    used internally, mainly to keep vcdimager_xml() from being too long.
    """
    menu = titleset.menu
    videos = titleset.videos
    # Add menu
    if menu:
        segment_items.add('segment-item', src=menu.filename,
                          id='seg-menu-%d' % ts_id)
        # Add menu to pbc
        selection = pbc.add('selection', id='select-menu-%d' % ts_id)
        selection.add('bsn', '1')
        # Link back to topmenu (not sure what'll happen if there isn't one...)
        selection.add('return', ref='select-topmenu')
        selection.add('loop', '0', jump_timing='immediate')
        selection.add('play-item', ref='seg-menu-%d' % ts_id)
        # Navigational links to titleset videos
        for video_id in range(len(videos)):
            selection.add('select',
                          ref='play-title-%d-%d' % (ts_id, video_id))
    # Add videos
    for video_id, video in enumerate(videos):
        # Add sequence items
        sequence_item = sequence_items.add('sequence-item')
        sequence_item.set(id='seq-title-%d-%d' % (ts_id, video_id),
                          src=video.filename)
        # Add chapter entries
        for chapterid, chapter in enumerate(video.chapters):
            sequence_item.add('entry', chapter, id='chapter-%d' % chapterid)
        # Add playlists to pbc
        playlist = pbc.add('playlist')
        playlist.set(id='play-title-%d-%d' % (ts_id, video_id))
        # Add prev/next links if appropriate
        if 0 <= video_id < len(videos):
            if 0 < video_id:
                playlist.add('prev',
                             ref='play-title-%d-%d' % (ts_id, video_id-1))
            if video_id < len(videos) - 1:
                playlist.add('next',
                             ref='play-title-%d-%d' % (ts_id, video_id+1))
        # Add a return link to the menu, if there is one
        if menu:
            playlist.add('return', ref='select-menu-%d' % ts_id)
        playlist.add('wait', '0')
        playlist.add('play-item', ref='seq-title-%d-%d' % (ts_id, video_id))


###
### Exported functions
###

def vcdimager_xml(disc):
    """Return the vcdimager XML string for the given Disc.
    """
    assert isinstance(disc, Disc)
    # XML header (will be added later)
    header = '<?xml version="1.0"?>\n'
    header += '<!DOCTYPE videocd PUBLIC "-//GNU/DTD VideoCD//EN"\n'
    header += '  "http://www.gnu.org/software/vcdimager/videocd.dtd">\n'
    # Root videocd element
    attributes = {
        'xmlns': 'http://www.gnu.org/software/vcdimager/1.0/',
        'class': disc.format}
    if disc.format == 'vcd':
        attributes['version'] = '2.0'
    else: # SVCD
        attributes['version'] = '1.0'
    root = xml.Element('videocd')
    root.set(attributes)
    # Add info block
    info = root.add('info')
    info.add('album-id', 'VIDEO_DISC')
    info.add('volume-count', '1')
    info.add('volume-number', '1')
    info.add('restriction', '0')
    # Add pvd block
    pvd = root.add('pvd')
    pvd.add('volume-id', 'VIDEO_DISC')
    pvd.add('system-id', 'CD-RTOS CD-BRIDGE')

    # Add segment, sequence, and pbc
    segment_items = root.add('segment-items')
    sequence_items = root.add('sequence-items')
    pbc = root.add('pbc')
    # Add topmenu
    if disc.topmenu:
        segment_items.add('segment-item', src=disc.topmenu.filename,
                          id='seg-topmenu')
        selection = pbc.add('selection', id='select-topmenu')
        selection.add('bsn', '1')
        selection.add('loop', '0', jump_timing='immediate')
        selection.add('play-item', ref='seg-topmenu')
    # Add titlesets
    for ts_id, titleset in enumerate(disc.titlesets):
        _add_titleset(titleset, ts_id, segment_items, sequence_items, pbc)
        # If we're doing a topmenu, add a link to the titleset menu
        if disc.topmenu:
            selection.add('select', ref='select-menu-%d' % ts_id)
    
    # Return XML with header prepended
    return header + str(root) + '\n'


def dvdauthor_xml(disc):
    """Return the dvdauthor XML string for the given Disc.
    """
    assert isinstance(disc, Disc)
    # Root dvdauthor element
    root = xml.Element('dvdauthor', dest=disc.name.replace(' ', '_'))
    vmgm = root.add('vmgm')
    # Add topmenu if present
    if disc.topmenu:
        menus = vmgm.add('menus')
        menus.add('video')
        pgc = menus.add('pgc', entry='title')
        vob = pgc.add('vob', file=disc.topmenu.filename)
        for index, titleset in enumerate(disc.titlesets):
            if titleset.menu:
                pgc.add('button', 'jump titleset %d menu;' % (index + 1))
            else:
                pgc.add('button', 'jump titleset %d;' % (index + 1))
    # Add each titleset
    for titleset in disc.titlesets:
        menu = titleset.menu
        ts = root.add('titleset')
        # Add menu if present
        if menu:
            menus = ts.add('menus')
            menus.add('video')
            pgc = menus.add('pgc', entry='root')
            vob = pgc.add('vob', file=menu.filename)
            for index, video in enumerate(titleset.videos):
                pgc.add('button', 'jump title %d;' % (index + 1))
            if disc.topmenu:
                pgc.add('button', 'jump vmgm menu;')
        titles = ts.add('titles')
        # Add a pgc for each video
        for video in titleset.videos:
            pgc = titles.add('pgc')
            vob = pgc.add('vob', file=video.filename)
            if video.chapters:
                vob.set(chapters=','.join(video.chapters))
            pgc.add('post', 'call menu;')
    # Return XML string
    return str(root) + '\n'


###
### Self-test
###

if __name__ == '__main__':
    menu1 = Menu('menu1.mpg')
    videos1 = [
        Video('video1.mpg'),
        Video('video2.mpg'),
        Video('video3.mpg')]
    menu2 = Menu('menu2.mpg')
    videos2 = [
        Video('video4.mpg'),
        Video('video5.mpg'),
        Video('video6.mpg')]
    titlesets = [
        Titleset(menu1, videos1),
        Titleset(menu2, videos2)]
    disc = Disc('dvd_test', 'dvd', 'ntsc', titlesets)

    print "dvdauthor XML example:"
    print dvdauthor_xml(disc)

    print "vcdimager XML example:"
    print vcdimager_xml(disc)



#! /usr/bin/env python
# standard.py

"""This module defines functions for retrieving information about multimedia
standards, including functions for determining the appropriate resolution,
video and audio codec, fps, and bitrates for a given format.
"""

__all__ = [\
    'abitrate',
    'acodec',
    'fps',
    'fpsratio',
    'resolution',
    'samprate',
    'scaling',
    'vbitrate',
    'vcodec']

import doctest

def resolution(format, tvsys):
    """Return the pixel resolution (x,y) for the given format and TV system.
    For example:

        >>> resolution('dvd', 'pal')
        (720, 576)
        >>> resolution('half-dvd', 'ntsc')
        (352, 480)
    """
    # Valid resolutions, indexed by format and tvsys
    valid_size = {\
        'vcd':
            {'pal': (352, 288), 'ntsc': (352, 240)},
        'dvd-vcd':
            {'pal': (352, 288), 'ntsc': (352, 240)},
        'svcd':
            {'pal': (480, 576), 'ntsc': (480, 480)},
        'half-dvd':
            {'pal': (352, 576), 'ntsc': (352, 480)},
        'dvd':
            {'pal': (720, 576), 'ntsc': (720, 480)}
        }
    return valid_size[format][tvsys]


def vcodec(format):
    """Return the video codec used by the given format. For example:
    
        >>> vcodec('vcd')
        'mpeg1'
        >>> vcodec('svcd')
        'mpeg2'
    """
    if format == 'vcd':
        return 'mpeg1'
    else:
        return 'mpeg2'


def acodec(format):
    """Return the audio codec (or codecs) supported by the given format. For
    example:
    
        >>> acodec('vcd')
        'mp2'
        >>> acodec('dvd')
        'ac3'
    """
    if format in ['vcd', 'svcd']:
        return 'mp2'
    else:
        return 'ac3'


def samprate(format):
    """Return the audio sampling rate used by the given format."""
    if format in ['vcd', 'svcd']:
        return 44100
    else:
        return 48000
    

def fps(tvsys):
    """Return the number of frames per second for the given TV system. For
    example:
    
        >>> print fps('ntsc')
        29.97
        >>> print fps('pal')
        25.0
    """
    # Valid frames per second, by TV system
    fps = {\
        'pal': 25.0,
        'ntsc': 29.97,
        'ntscfilm': 23.976,
        }
    return fps[tvsys]


def fpsratio(tvsys):
    """Return the number of frames per second for the given TV system, in
    ratio form. For example:
    
        >>> fpsratio('ntsc')
        '30000:1001'
        >>> fpsratio('pal')
        '25:1'
    """
    # Valid frames per second, by TV system
    fps = {\
        'pal': '25:1',
        'ntsc': '30000:1001',
        'ntscfilm': '24000:1001',
        }
    return fps[tvsys]


def vbitrate(format):
    """Return the range (min, max) of valid video bitrates (in kilobits per
    second) for the given format. min and max are the same for constant-bitrate
    formats.
    """
    # Valid video bitrates, indexed by format
    valid_bitrates = {\
        'vcd': (1150, 1150),
        'svcd': (0, 2600),
        'dvd-vcd': (0, 9800),
        'half-dvd': (0, 9800),
        'dvd': (0, 9800)
        }
    return valid_bitrates[format]


def abitrate(format):
    """Return the range (min, max) of valid audio bitrates (in kilobits per
    second) for the given format. min and max are the same for constant-bitrate
    formats.
    """
    if format == 'vcd':
        return (224, 224)
    elif format == 'svcd':
        return (32, 384)
    else:
        return (32, 1536)


if __name__ == '__main__':
    doctest.testmod(verbose=True)



#! /usr/bin/env python
# support.py

"""Supporting classes for metagui"""

__all__ = [
    'ListVar',
    'ScrollList',
    'DragList',
    'ComboBox',
    'FontChooser',
    'ConfigWindow',
    'ScrolledWindow',
    'Style']

import os
import Tkinter as tk
import tkSimpleDialog

### --------------------------------------------------------------------

class ListVar (tk.Variable):
    """A tk Variable suitable for associating with Listboxes.
    """
    def __init__(self, master=None, items=None):
        """Create a ListVar with a given master and initial list of items.
        """
        tk.Variable.__init__(self, master)
        if items:
            self.set(items)
    
    def __getitem__(self, index):
        """Get a list value using list-index syntax (listvar[index]).
        """
        return self.get()[index]

    def __setitem__(self, index, value):
        """Set a list value using list-index syntax (listvar[index] = value).
        """
        current = self.get()
        current[index] = value
        self.set(current)

    def get(self):
        return list(tk.Variable.get(self))
    
    def set(self, new_list):
        tk.Variable.set(self, tuple(new_list))

    def remove(self, item):
        """Remove the item from the list, if it exists.
        """
        items = self.get()
        items.remove(item)
        self.set(items)

    def append(self, item):
        """Append an item to the list.
        """
        items = self.get()
        self.set(items + [item])

### --------------------------------------------------------------------

class ScrollList (tk.Frame):
    """A Listbox with a scrollbar.

    Similar to a tk.Listbox, a ScrollList shows a list of items. tk.Variables
    may be associated with both the list of available choices, and the one
    that is currently chosen/highlighted.
    """
    def __init__(self, master=None, choices=None,
                 chosen=None, command=None):
        """Create a ScrollList widget.
        
            master:   Tkinter widget that will contain the ScrollList
            choices:  ListVar or Python list of choices to show in listbox
            chosen:   Tk StringVar to store currently selected choice in
            command:  Function to call when a list item is clicked
        """
        tk.Frame.__init__(self, master)
        if type(choices) == list:
            choices = ListVar(self, choices)
        self.choices = choices or ListVar()
        self.chosen = chosen or tk.StringVar()
        self.command = command
        self.curindex = 0
        self.linked = None
        # Draw listbox and scrollbar
        self.scrollbar = tk.Scrollbar(self, orient='vertical',
                                      command=self.scroll)
        self.listbox = tk.Listbox(self, width=30, listvariable=self.choices,
                                  yscrollcommand=self.scrollbar.set)
        self.listbox.pack(side='left', fill='both', expand=True)
        self.scrollbar.pack(side='left', fill='y')
        self.listbox.bind('<Button-1>', self.select)

    def add(self, *values):
        """Add the given values to the list.
        """
        for value in values:
            self.listbox.insert('end', value)

    def insert(self, index, *values):
        """Insert values at a given index.
        """
        self.listbox.insert(index, *values)

    def delete(self, first, last=None):
        """Delete values in a given index range (first, last), not including
        last itself. If last is None, delete only the item at first index.
        """
        self.listbox.delete(first, last=None)

    def scroll(self, *args):
        """Event handler when the list is scrolled.
        """
        apply(self.listbox.yview, args)
        if self.linked:
            apply(self.linked.listbox.yview, args)

    def select(self, event):
        """Event handler when an item in the list is selected.
        """
        self.curindex = self.listbox.nearest(event.y)
        self.chosen.set(self.listbox.get(self.curindex))

    def get(self):
        """Return a list of all entries in the list.
        """
        return self.choices.get()
    
    def set(self, values):
        """Set the list values to those given.
        """
        self.choices.set(values)

    def swap(self, index_a, index_b):
        """Swap the element at index_a with the one at index_b.
        """
        item_a = self.choices[index_a]
        item_b = self.choices[index_b]
        self.choices[index_a] = item_b
        self.choices[index_b] = item_a

    def link(self, scrolllist):
        """Link this list to another, so they scroll in unison."""
        if not isinstance(scrolllist, ScrollList):
            raise TypeError("Can only link to a ScrollList.")
        self.linked = scrolllist

### --------------------------------------------------------------------

class DragList (ScrollList):
    """A scrollable listbox with drag-and-drop support"""
    def __init__(self, master=None, choices=None,
                 chosen=None, command=None):
        """Create a DragList widget.
        
            master:   Tkinter widget that will contain the DragList
            choices:  ListVar or Python list of choices to show in listbox
            chosen:   Tk StringVar to store currently selected choice in
            command:  Function to call when a list item is clicked
        """
        ScrollList.__init__(self, master, choices, chosen, command)
        # Add bindings for drag/drop
        self.listbox.bind('<Button-1>', self.select)
        self.listbox.bind('<B1-Motion>', self.drag)
        self.listbox.bind('<ButtonRelease-1>', self.drop)

    def select(self, event):
        """Event handler when an item in the list is selected.
        """
        # Set currently selected item and change the cursor to a double-arrow
        ScrollList.select(self, event)
        if self.linked:
            self.linked.curindex = self.curindex
        self.config(cursor="double_arrow")
    
    def drag(self, event):
        """Event handler when an item in the list is dragged.
        """
        # If item is dragged to a new location, delete/insert
        loc = self.listbox.nearest(event.y)
        if loc != self.curindex:
            #item = self.listbox.get(self.curindex)
            #self.listbox.delete(self.curindex)
            #self.listbox.insert(loc, item)
            self.swap(self.curindex, loc)
            # Drag in linked listbox
            if self.linked:
                self.linked.swap(self.curindex, loc)
                self.linked.curindex = loc
            self.curindex = loc
    
    def drop(self, event):
        """Event handler when an item in the list is "dropped".
        """
        # Change the mouse cursor back to the default arrow.
        self.config(cursor="")

### --------------------------------------------------------------------

class ComboBox (tk.Frame):
    """A dropdown menu with several choices.
    """
    def __init__(self, master, choices=None,
                 variable=None, command=None):
        """Create a ComboBox.
            
            master:     Tk Widget that will contain the ComboBox
            choices:    ListVar or Python list of available choices
            variable:   Tk StringVar to store currently selected choice in
            command:    Function to call after an item in the list is chosen
        """
        tk.Frame.__init__(self, master)
        if type(choices) == list:
            choices = ListVar(self, choices)
        self.choices = choices or ListVar()
        self.variable = variable or tk.StringVar()
        self.command = command
        self._draw()

    def _draw(self):
        """Draw and configure contained widgets.
        """
        # Text and button
        self.text = tk.Entry(self, textvariable=self.variable)
        self.text.grid(row=0, column=0)
        self.button = tk.Button(self, text="<", command=self.open)
        self.button.grid(row=0, column=1)

        # Dropdown list, displayed when button is clicked
        self.dropdown = tk.Toplevel(self)
        # Don't draw window manager frame around the dropdown
        self.dropdown.wm_overrideredirect(1)
        # Hide until later
        self.dropdown.withdraw()

        # List of choices
        self.chooser = tk.Listbox(self.dropdown, background='white',
                                  listvariable=self.choices)
        #for choice in self.choices:
        #    self.chooser.insert('end', choice)
        self.chooser.bind('<Button-1>', self.choose)
        self.chooser.grid()
    
    def open(self):
        """Open/close a panel showing the list of choices.
        """
        if self.dropdown.winfo_viewable():
            self.dropdown.withdraw()
        else:
            # Align dropdown list with the text box
            x = self.text.winfo_rootx()
            y = self.text.winfo_rooty()
            self.dropdown.wm_geometry("+%d+%d" % (x, y))
            # Show list
            self.dropdown.deiconify()
    
    def choose(self, event=None):
        """Make a selection from the list, and set the variable.
        """
        self.curindex = self.chooser.nearest(event.y)
        self.variable.set(self.chooser.get(self.curindex))
        self.dropdown.withdraw()
        # Callback, if any
        if self.command:
            self.command()

### --------------------------------------------------------------------

class FontChooser (tkSimpleDialog.Dialog):
    """A widget for choosing a font"""
    def __init__(self, master=None):
        tkSimpleDialog.Dialog.__init__(self, master, "Font chooser")

    def get_fonts(self):
        """Return a list of font names available in ImageMagick.
        """
        find = "convert -list type | sed '/Path/,/---/d' | awk '{print $1}'"
        return [line.rstrip('\n') for line in os.popen(find).readlines()]

    def body(self, master):
        """Draw widgets inside the Dialog, and return the widget that should
        have the initial focus. Called by the Dialog base class constructor.
        """
        tk.Label(master, text="Available fonts").pack(side='top')
        self.fontlist = ScrollList(master, self.get_fonts())
        self.fontlist.pack(side='top', fill='both', expand=True)
        # Return widget with initial focus
        return self.fontlist
    
    def apply(self):
        """Set the selected font.
        """
        self.result = self.fontlist.chosen.get()

### --------------------------------------------------------------------
from ConfigParser import ConfigParser

class Style:
    """Generic widget style definitions."""
    def __init__(self,
                 bgcolor='white',
                 fgcolor='grey',
                 textcolor='black',
                 font=('Helvetica', 10, 'normal'),
                 relief='groove'):
        self.bgcolor = bgcolor
        self.fgcolor = fgcolor
        self.textcolor = textcolor
        self.font = font
        self.relief = relief

    def apply(self, root):
        """Apply the current style to the given Tkinter root window."""
        assert isinstance(root, tk.Tk)
        root.option_clear()
        # Font
        root.option_add("*font", self.font)
        # Background color
        root.option_add("*Scale.troughColor", self.bgcolor)
        root.option_add("*Spinbox.background", self.bgcolor)
        root.option_add("*Entry.background", self.bgcolor)
        root.option_add("*Listbox.background", self.bgcolor)
        # Button colors
        root.option_add("*Radiobutton.selectColor", "#8888FF")
        root.option_add("*Checkbutton.selectColor", "#8888FF")
        # Relief
        root.option_add("*Entry.relief", self.relief)
        root.option_add("*Spinbox.relief", self.relief)
        root.option_add("*Listbox.relief", self.relief)
        root.option_add("*Button.relief", self.relief)
        root.option_add("*Menu.relief", self.relief)
        # Mouse-over effects
        root.option_add("*Button.overRelief", 'raised')
        root.option_add("*Checkbutton.overRelief", 'raised')
        root.option_add("*Radiobutton.overRelief", 'raised')

    def save(self, filename):
        """Save the current style settings to an .ini-formatted config file.
        """
        # Load existing config file
        config = ConfigParser()
        config.read(filename)

        # Save font settings
        if 'font' not in config.sections():
            config.add_section('font')
        family, size, style = self.font
        config.set('font', 'family', family)
        config.set('font', 'size', size)
        config.set('font', 'style', style)
        
        # TODO: Save other style settings
        
        # Yuck...
        dirname, fname = os.path.split(filename)
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        outfile = open(filename, 'w')
        config.write(outfile)
        outfile.close()

    def load(self, filename):
        """Load style settings from an .ini-formatted config file.
        """
        config = ConfigParser()
        config.read(filename)
        font = dict(config.items('font'))
        self.font = (font['family'], int(font['size']), font['style'])

### --------------------------------------------------------------------
from tkMessageBox import showinfo

class ConfigWindow (tkSimpleDialog.Dialog):
    """Configuration settings dialog box.
    """
    def __init__(self, master=None, style=None):
        """Create and display a configuration window.
        
            inifile:  An .ini-style file to load/save settings from
        """
        self.style = style or Style()
        tkSimpleDialog.Dialog.__init__(self, master, "Configuration")

    def body(self, master):
        """Draw widgets inside the Dialog, and return the widget that should
        have the initial focus. Called by the Dialog base class constructor.
        """
        # Font family
        tk.Label(master, text="Font family").pack(side='top')
        self.fontfamily = ComboBox(master,
                                   choices=['Helvetica', 'Courier', 'Times'])
        self.fontfamily.pack(side='top', fill='both', expand=True)
        # Font size
        tk.Label(master, text="Font size").pack(side='top')
        self.fontsize = ComboBox(master, choices=[8, 10, 12, 15, 18, 24])
        self.fontsize.pack(side='top')
        # Font style
        tk.Label(master, text="Font style").pack(side='top')
        self.fontstyle = ComboBox(master, choices=['normal', 'bold'])
        self.fontstyle.pack(side='top')
        # Use initial values loaded from .ini file
        family, size, style = self.style.font
        self.fontfamily.variable.set(family)
        self.fontsize.variable.set(size)
        self.fontstyle.variable.set(style)
        # Return widget with initial focus
        return self.fontfamily

    def apply(self):
        """Apply the selected configuration settings.
        """
        self.style.font = (self.fontfamily.variable.get(),
                           self.fontsize.variable.get(),
                           self.fontstyle.variable.get())
        self.result = self.style

### --------------------------------------------------------------------

class ScrolledWindow (tk.Tk):
    """A top-level window with scrollbars.
    
    To use as a container for other widgets, do:
    
        window = ScrolledWindow()
        button = tk.Button(window.frame, text="Click me", ...)
        entry = tk.Entry(window.frame, ...)
        window.draw()
    
    That is, use window.frame as the master of child widgets, instead of
    window itself. (TODO: Eliminate this requirement.)
    """
    def __init__(self, width, height):
        tk.Tk.__init__(self)
        self.width = width
        self.height = height
        # Frame inside canvas, fills all available space
        self.canvas = tk.Canvas(self, width=self.width, height=self.height)
        self.frame = tk.Frame(self.canvas)
        # Grid fills all available window space
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)

    def draw(self):
        """Draw the scrollbars and container frame.
        """
        # Put the container frame in the Canvas
        self.canvas.create_window(0, 0, window=self.frame, anchor='nw')
        # Canvas scrollable area
        self.canvas.configure(scrollregion=(0, 0, self.width, self.height))
        self.canvas.grid(row=0, column=0, sticky='nsew')
        self.draw_scrollbars()

    def draw_scrollbars(self):
        # Attach scrollbars to the Canvas
        h_scroll = tk.Scrollbar(self, orient='horizontal',
                                command=self.canvas.xview)
        v_scroll = tk.Scrollbar(self, orient='vertical',
                                command=self.canvas.yview)
        h_scroll.grid(row=1, column=0, sticky='we')
        v_scroll.grid(row=0, column=1, sticky='ns')
        self.canvas.configure(xscrollcommand=h_scroll.set,
                              yscrollcommand=v_scroll.set)

### --------------------------------------------------------------------



#! /usr/bin/env python
# deps.py

"""This module is for verifying libtovid's run-time dependencies.
For example:

    deps.require(deps.core)

will look for all of tovid's core dependencies. If any cannot be found, the
missing ones are printed and an exception is raised. Run deps.py from a prompt
for further examples.

In practice:

    try:
        deps.require(deps.core)
    except deps.MissingError:
        print "Exiting..."
        sys.exit(1)

deps.core is an internal dictionary of tovid's core dependencies, where the
keys are the names, and the values are brief descriptions with URLs.

Provided dependency dictionaries:
    core      -- grep/sed/md5sum, mplayer, mencoder, mpjpegtools, ffmpeg
    magick    -- composite, convert
    dvd       -- spumux, dvdauthor, growisofs
    vcd       -- vcdxbuild, cdrdao
    transcode -- tcprobe, tcrequant
    plugin    -- sox, normalize
    all       -- ALL dependencies above
    
If you don't want to use a provided dictionary, you can specify individual
program names to look for:

    deps.require("more less cat")

require also provides ways to print custom URLs and help when it cannot find
dependencies. See help(deps.require) or keep reading.
"""

__all__ = [
    'which',
    'require']

from subprocess import Popen, PIPE
import textwrap
from libtovid.output import red

###
### Exceptions
###

class DepError(Exception): pass
class InputError(DepError): pass
class MissingError(DepError): pass

###
### Module data
###

# Dictionary format: {"name": "description (url)"}
all = {}

core = { 
    "grep":         "a GNU utility (www.gnu.org/software/grep)",
    "sed":          "a GNU utility (directory.fsf.org/GNU/software/sed.html)",
    "md5sum":       "a GNU utility (www.gnu.org/software/coreutils)",
    "mplayer":      "part of mplayer (www.mplayerhq.hu)",
    "mencoder":     "part of mplayer (www.mplayerhq.hu)",
    "mplex":        "part of mjpegtools (mjpeg.sf.net)",
    "mpeg2enc":     "part of mjpegtools (mjpeg.sf.net)",
    "yuvfps":       "part of mjpegtools (mjpeg.sf.net)",
    "yuvdenoise":   "part of mjpegtools (mjpeg.sf.net)",
    "ppmtoy4m":     "part of mjpegtools (mjpeg.sf.net)",
    "mp2enc":       "part of mjpegtools (mjpeg.sf.net)",
    "jpeg2yuv":     "part of mjpegtools (mjpeg.sf.net)",
    "ffmpeg":       "a video encoding utility (ffmpeg.mplayerhq.hu)"      }
all.update(core)

magick = {
    "composite":    "part of ImageMagick (www.imagemagick.org)",
    "convert":      "part of ImageMagick (www.imagemagick.org)"     }
all.update(magick)

dvd = {
    "spumux":       "part of dvdauthor (dvdauthor.sf.net)",
    "dvdauthor":    "part of dvdauthor (dvdauthor.sf.net)",
    "growisofs":    "part of dvd+rw-tools (fy.chalmers.se/~appro/linux/DVD+RW)"}
all.update(dvd)

vcd = {
    "vcdxbuild":    "part of vcdimager (www.vcdimager.org)",
    "cdrdao":       "a cd burning application (cdrdao.sf.net)"   }
all.update(vcd)

transcode = {
    "tcprobe":      "part of transcode (www.transcoding.org)",
    "tcrequant":    "part of transcode (www.transcoding.org)"   }
all.update(transcode)

plugin = {
    "sox":          "swiss army knife for sound (sox.sf.net)",
    "normalize":    "wave gain and normalization (normalize.nongnu.org)" }
all.update(plugin)

__missing_dependency_message = \
"""Please install the above MISSING dependencies and try again. See
http://tovid.wikia.com/wiki/Tovid_dependencies or http://tovid.org
for more help.
"""

###
### Exported functions
###

def which(executable):
    """Locate the given executable program name on the current path.
    If found, return the full pathname; otherwise, return the empty string.
    """
    proc = Popen(['which', executable], stdout=PIPE, stderr=PIPE)
    return proc.stdout.read().rstrip('\n')

def require(deps, 
            help="You need these to finish what you were doing.",
            description="oops"):

    """Assert that one or more dependencies exist on the system, raise
    a 'MissingError' exception if not.
    
        deps:        Names of dependencies to assert. May be a single name,
                     a python list or space-separated string list of names,
                     or a dictionary of the form {depname: description}.
        help:        A help message about why the dependencies are needed
        description: A short description of the dep, and its homepage URL
                     (ignored if deps is a dictionary).

    If a given dependency name is found in one of the internal dictionaries,
    (core, magick, etc.), its description is taken from there.

    Examples:
        require(all)
            Look for ALL dependencies that are defined internally.

        require(core, "You are missing CORE dependencies!")
            Use the core dictionary to determine which dependencies to 
            look for. Print the help message if any dependency is missing.

        require("xine", "You can't preview!", "a video player (xinehq.de)")
            Check for xine and print the custom help and URL messages
            if missing.

        require("mplayer mencoder", url="a video player (mplayerhq.hu)")
            Look for mplayer and mencoder. Use the default help message,
            but print a custom URL.

    """
    # Get the deps to check for: dictionary.keys(), or a list
    if type(deps) == dict:
        deps = deps.keys()
    elif type(deps) == str:
        deps = deps.split(' ')
    elif type(deps) != list:
        raise InputError, "%s is not dictionary, list, or string!" % str(deps)

    # Find the missing dependencies
    have_deps = True
    for dep in deps:
        if which(dep) == '':
            if type(deps) == dict:
                description = deps[dep]
            elif dep in all:
                description = all[dep]
            # Tell the user which dependencies they're missing
            print red("MISSING:") + " %s, %s" % (dep, description)
            have_deps = False

    # Having reported the missing dependencies, print the help message and quit
    if not have_deps:
        print "\n", textwrap.fill(help + ' ' + __missing_dependency_message, 79)
        raise MissingError, "Cannot find required dependencies!"


###
### Main
###

if __name__ == "__main__":
    trials = [  
        (core,        "You are missing CORE dependencies!", "custom url 1"),
        (core,        "You are missing CORE dependencies!", "custom url 2"),
        ("grp md5sm", "You're missing some vowels!",        "custom url 4"),
        ("foo",       "heh, you don't have foo?!",          "see foo.org")   ]

    for dep, help, url in trials:
        try:
            print "================================================"
            print "Starting test..."
            require(dep, help, url)
        except MissingError:
            print "<< Would exit here. >>\n"
        else:
            print "Test passed!\n"

